/**
   Bojan Nikolic <b.nikolic@mrao.cam.ac.uk>, <bojan@bnikolic.co.uk>
   Initial version January 2010.
   Maintained by ESO since 2013.

   This file is part of LibAIR and is licensed under GNU Public
   License Version 2

   \file dtdlcoeffs.cpp
   Renamed dtdlcoeffs.cc 2023

*/

#include <algorithm> 
#include <numeric>

#include <cmath>
#include <iostream>
#include <stdio.h>            /*** for sprintf(...) ***/

#include "dtdlcoeffs.h"
#include "almaresults.h"

namespace LibAIR2 {

  dTdLCoeffsBase::dTdLCoeffsBase()
  {
    std::array<double, 4> _chmask={1,1,1,1};
    chmask=_chmask;
  }

  dTdLCoeffsBase::~dTdLCoeffsBase()
  {
  }

  dTdLCoeffsSingle::dTdLCoeffsSingle(const std::vector<double> &c,
				     const std::vector<double> &e):
    c(c),
    c2(4, 0.0),
    e(e)
  {
  }

  dTdLCoeffsSingle::dTdLCoeffsSingle(const  ALMAResBase &r):
    c(4),
    c2(4, 0.0),
    e(4)
  {
    for(size_t i=0; i<4; ++i)
    {
      c[i]=r.dTdL[i];
      e[i]=r.dTdL_err[i];
    }
  }

  void dTdLCoeffsSingle::get(size_t i,
			     double time,
			     double el,
			     std::vector<double> &res,
			     std::vector<double> &res2) const
  {
    (void) i; // use not yet implemented
    (void) time;
    (void) el;
    res.resize(c.size());
    std::copy(c.begin(), 
	      c.end(),
	      res.begin());

    for(size_t j=0; j<4; ++j)
      res[j]*=chmask[j];

    res2.resize(c2.size());
    std::copy(c2.begin(), 
	      c2.end(),
	      res2.begin());
	      
  }

  void dTdLCoeffsSingle::print(std::ostream &os)
  {
    os<<"Coefficients: ";
    for(size_t i=0; i<4; ++i)
      os<<c[i]<<", ";
    os<<std::endl;
  }

  void dTdLCoeffsSingle::repr(std::vector<double> &res,
			      std::vector<double> &err) const
  {
    res=c;
    err=e;
  }


  bool dTdLCoeffsSingle::isnan(void) const
  {
    return std::accumulate(c.begin(),c.end(),false,[](bool acc, double v) { return acc || std::isnan(v); });
  }

  dTdLCoeffsIndiv::dTdLCoeffsIndiv(const coeff_t &c):
    coeff(c)
  {
  }

  dTdLCoeffsIndiv::dTdLCoeffsIndiv(size_t nAnts):
    coeff(4,nAnts,3,0.)
  {
  }

  void dTdLCoeffsIndiv::set(size_t i,
			    const std::vector<double> &c,
			    const std::vector<double> &e)
  {
    for(size_t k=0; k<4; ++k)
    {
      coeff(k,i,0)=c[k];
      coeff(k,i,1)=e[k];
    }
  }

  void dTdLCoeffsIndiv::get(size_t i,
			    double time,
			    double el,
			    std::vector<double> &res,
			    std::vector<double> &res2) const
  {
    (void) time; // use not yet implemented
    (void) el;

    const size_t nc=4; //coeff.shape()[0];
    res.resize(nc);
    res2.resize(nc);
    for(size_t k=0; k<nc; ++k)
    {
      res[k]=coeff(k,i,0)*chmask[k];
      res2[k]=coeff(k,i,2);
    }
  }

  void dTdLCoeffsIndiv::print(std::ostream &os)
  {
    os<<"Coefficients: ";
    for(size_t i=0; i<4; ++i)
      os<<coeff(i,0,0)<<", ";
    os<<std::endl;
  }

  void dTdLCoeffsIndiv::repr(std::vector<double> &res,
			     std::vector<double> &err) const
  {
    const size_t nc=4; //coeff.shape()[0];
    res.resize(nc);
    err.resize(nc);
    for(size_t k=0; k<nc; ++k)
    {
      res[k]=coeff(k,0,0);
      err[k]=coeff(k,0,1);
    }
  }

  bool dTdLCoeffsIndiv::isnan(void) const
  {
    // return True if one of the elements of coeff is a NaN
    if (coeff.contiguousStorage()) {
      return std::accumulate(coeff.cbegin(), 
			     coeff.cend(),
			     false,
			     [](bool acc, double v) { return acc || std::isnan(v); }
			     );
    } else {
      return std::accumulate(coeff.begin(), 
			     coeff.end(),
			     false,
			     [](bool acc, double v) { return acc || std::isnan(v); }
			     );
    }
    
    //return std::accumulate(coeff.origin(), 
    //			   coeff.origin()+(coeff.shape()[0]*coeff.shape()[1]*coeff.shape()[2]),
    //			   false,
    //			   [](bool acc, double v) { return acc || std::isnan(v); }
    //			   );
  }


  dTdLCoeffsSingleInterpolated::dTdLCoeffsSingleInterpolated()
  {
  }

  void dTdLCoeffsSingleInterpolated::insert(double time,
					    const std::array<double, 4> &coeffs,
					    const std::array<double, 4> &err)
  {
    ret_t t;
    t.time=time;
    t.coeffs=coeffs;
    t.err=err;
    retrievals.insert(t);
  }

  void dTdLCoeffsSingleInterpolated::get(size_t i,
					 double time,
					 double el,
					 std::vector<double> &res,
					 std::vector<double> &c2) const
  {
    (void) i; // use not yet implemented
    (void) el;

    if (retrievals.size() == 0) 
    {
      throw std::runtime_error("No retrievals have been entered.");
    }

    res.resize(4);
    c2.resize(4);
    if (time<retrievals.begin()->time)
    {
      // request is before the first retrieval
      std::copy(retrievals.begin()->coeffs.begin(), 
		retrievals.begin()->coeffs.end(),
		res.begin());      
      for(size_t j=0; j<4; ++j)
	res[j]*=chmask[j];
      std::copy(retrievals.begin()->c2.begin(), 
		retrievals.begin()->c2.end(),
		c2.begin());      
      return;
    }
    std::set<ret_t>::const_iterator prev;
    std::set<ret_t>::const_iterator next;

    for(next=retrievals.begin();
	next !=retrievals.end() and time >= next->time;
	++next)
    {
    }

    if (next ==retrievals.end())
    {
      // requested time is after the last retrieval
      std::set<ret_t>::reverse_iterator last=retrievals.rbegin();
      std::copy(last->coeffs.begin(), 
		last->coeffs.end(),
		res.begin());         
      for(size_t j=0; j<4; ++j)
	res[j]*=chmask[j];
      std::copy(last->c2.begin(), 
		last->c2.end(),
		c2.begin());            
      return;
    }

    prev=next;
    prev--;


    //Finally if not before beginning or after end, interpolate
    //linearly
    
    double c_f=(time - prev->time)/(next->time - prev->time);
    double c_s=(next->time - time)/(next->time - prev->time);
    
    for(size_t k=0; k<4; ++k)
    {
      res[k]= c_s*(prev->coeffs[k]) + c_f*(next->coeffs[k]);
      res[k]*= chmask[k];
      c2[k]= c_s*(prev->c2[k]) + c_f*(next->c2[k]);
    }
  }

  void dTdLCoeffsSingleInterpolated::print(std::ostream &os)
  {
    for(std::set<ret_t>::const_iterator i=retrievals.begin();
	i!=retrievals.end();
	++i)
    {
      char tstr[120];
      sprintf(tstr, "Retrieval at %f is: [", i->time); 
      os << tstr;
      for(size_t j=0; j<4; ++j)
      {
	os << i->coeffs[j] <<",";
      }
      os << "]"
	 << std::endl;
    }
  }

  void dTdLCoeffsSingleInterpolated::repr(std::vector<double> &res,
					  std::vector<double> &err) const
  {
    std::set<ret_t>::const_iterator i=retrievals.begin();
    size_t j;
    for(j=0; j < retrievals.size()/2;++j)
    {
      ++i;
    };
    while(j < retrievals.size()-1) // make sure we have not hit an invalid coefficent
    {
      if(i->coeffs[0]>0.)
      {
	break;
      }
      ++j;
      ++i;
    };

    res.resize(4);
    err.resize(4);
    for(j=0; j<4; ++j)
    {
      res[j]=i->coeffs[j];
      err[j]=i->err[j];
    }
  }

  bool dTdLCoeffsSingleInterpolated::isnan() const
  {
    for(std::set<ret_t>::const_iterator i=retrievals.begin(); 
	i != retrievals.end();
	++i)
    {
      for(size_t j=0; j<4; ++j)
      {
	if ( std::isnan(i->coeffs[j]) || (std::isnan(i->err[j])) ||  std::isnan(i->c2[j]) )
	{
	  return true;
	}
      }
    }
    return false;
  }

}


