//# MSCache.cc: Specialized PlotMSCache for filling MSs
//# Copyright (C) 2009
//# Associated Universities, Inc. Washington DC, USA.
//#
//# This library is free software; you can redistribute it and/or modify it
//# under the terms of the GNU Library General Public License as published by
//# the Free Software Foundation; either version 2 of the License, or (at your
//# option) any later version.
//#
//# This library is distributed in the hope that it will be useful, but WITHOUT
//# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
//# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Library General Public
//# License for more details.
//#
//# You should have received a copy of the GNU Library General Public License
//# along with this library; if not, write to the Free Software Foundation,
//# Inc., 675 Massachusetts Ave, Cambridge, MA 02139, USA.
//#
//# Correspondence concerning AIPS++ should be addressed as follows:
//#        Internet email: casa-feedback@nrao.edu.
//#        Postal address: AIPS++ Project Office
//#                        National Radio Astronomy Observatory
//#                        520 Edgemont Road
//#                        Charlottesville, VA 22903-2475 USA
//#
//# $Id: $
#include <plotms/Data/MSCache.h>
#include <plotms/Data/PlotMSIndexer.h>
#include <plotms/Data/PlotMSAtm.h>

#include <casa/OS/Timer.h>
#include <casa/OS/Memory.h>
#include <casa/OS/Path.h>
#include <casa/Quanta/MVTime.h>
#include <casa/System/Aipsrc.h>
#include <casa/Utilities/Sort.h>
#include <casa/Arrays/ArrayMath.h>
#include <tables/Tables/ScalarColumn.h>
#include <lattices/Lattices/ArrayLattice.h>
#include <lattices/LatticeMath/LatticeFFT.h>
#include <measures/Measures/MDirection.h>
#include <measures/Measures/MPosition.h>
#include <measures/Measures/MFrequency.h>
#include <measures/Measures/Stokes.h>
#include <ms/MeasurementSets/MSAntennaColumns.h>
#include <ms/MeasurementSets/MSColumns.h>
#include <ms/MeasurementSets/MeasurementSet.h>
#include <ms/MSOper/MSMetaData.h>
#include <scimath/Mathematics/FFTServer.h>
#include <tables/Tables/Table.h>

#include <msvis/MSVis/VisSet.h>
#include <msvis/MSVis/VisBuffer2.h>
#include <mstransform/MSTransform/MSTransformIteratorFactory.h>
#include <mstransform/TVI/PointingInterpolationTVI.h>

#include <plotms/Data/PlotMSVBAverager.h>
#include <plotms/Data/MSCacheVolMeter.h>
#include <plotms/PlotMS/PlotMS.h>
#include <plotms/Plots/PlotMSPlotParameterGroups.h>

#include <casa/Logging/LogMessage.h>
#include <casa/Logging/LogSink.h>
#include <casa/Logging/LogFilter.h>

#include <ctime>

using namespace casacore;
namespace casa {

MSCache::MSCache(PlotMSApp* parent):
	PlotMSCacheBase(parent)
{
	ephemerisAvailable = false;
	vi_p = NULL;
	vm_ = NULL;
}

MSCache::~MSCache() {}

String MSCache::polname(Int ipol) {
	return Stokes::name(Stokes::type(ipol));
}


void MSCache::loadIt(vector<PMS::Axis>& loadAxes,
		vector<PMS::DataColumn>& loadData,
		ThreadCommunication* thread) {

	// process selected columns
	dataColumn_ = getDataColumn(loadAxes, loadData);

	// Get strings stored in MS 
	Table::TableOption tabopt(Table::Old);
	MeasurementSet* inputMS = new MeasurementSet(filename_, TableLock(TableLock::AutoLocking), tabopt);
	getNamesFromMS(*inputMS);

	// If plotting antennas directions, verify that feature is supported for the input MS
	auto loadAntDir = std::find_if(loadAxes.begin(),loadAxes.end(),PMS::axisIsRaDec) != loadAxes.end();
	if (loadAntDir) {
		auto ok = pointingsPlotSupported(inputMS);
		if (not ok){
			String errMsg("Plotting antennas directions is irrelevant or not supported for this MS.");
			logWarn(PMS::LOG_ORIGIN_LOAD_CACHE, errMsg);
			delete inputMS;
			loadError(errMsg);
		}
	}

	// Apply selections to MS to create selection MS and channel/correlation selections
	Vector<Vector<Slice> > chansel;
	Vector<Vector<Slice> > corrsel;
	MeasurementSet* selMS = new MeasurementSet();
	bool needStokesConverter(false);
	try {
		// get chansel, corrsel
		selection_.apply(*inputMS, *selMS, chansel, corrsel);
	} catch(const AipsError& log) {
		// improper selection can cause exception
		auto error = log.getMesg();
		bool selection_error(true);

		if (error.startsWith("Poln. expression error") ||
			(error.startsWith("MSSelectionNullSelection") &&
			!selection_.corr().empty())) {
			// Possibly a computed Stokes error; try it without correlation selection
			auto corr_selection = selection_.corr();
			bool is_corr_selection_error(false);

			try {
				// Reset corr selection and redo to check for other failures
				selection_.setCorr("");
				selection_.apply(*inputMS, *selMS, chansel, corrsel);
				is_corr_selection_error = true;
			} catch (AipsError& err) {
				// Selection without corr also failed
			}

			if (is_corr_selection_error) {
				Vector<String> corrs;
				corrs = split(corr_selection, ',', corrs);

				// Make list of valid Stokes types (stokes_out_) selected
				size_t ncorrs = corrs.size();
				stokes_out_.resize(ncorrs);
				bool is_valid_corr_selection = true;

				for (size_t i = 0; i < ncorrs; ++i) {
					// Convert each selection string into stokes type
					corrs(i).trim();
					auto stokes_type = casacore::Stokes::type(corrs(i));
					if (stokes_type == casacore::Stokes::Undefined) {
						// Invalid selection, quit.
						stokes_out_.resize();
						is_valid_corr_selection = false;
						break;
					} else {
						// Add to list
						stokes_out_[i] = stokes_type;

						if (stokes_type == casacore::Stokes::PFtotal ||
							stokes_type == casacore::Stokes::PFlinear) {
							// For error message, if div by I results in inf
							polnFraction_ = true;
						}
					}
				}

				if (is_valid_corr_selection) {
					// Do not log error and quit
					selection_error = false;

					// Check whether data or weight axis is selected (needs converter)
					for(unsigned int i = 0; i < loadAxes.size(); i++) {
						if (PMS::axisIsData(loadAxes[i]) || PMS::axisIsWeight(loadAxes[i])) {
							needStokesConverter = true;
							break;
						}
					}
				}
			}
		}

		if (selection_error) {
			delete inputMS;
			delete selMS;
			loadError(error);
		}
	}

	// Make volume meter for countChunks to estimate memory requirements
	vm_ = new MSCacheVolMeter(*selMS, averaging_, chansel, corrsel);

	// Load Page Header Cache
	loadPageHeaderCache(*selMS);

	delete inputMS;
	delete selMS;
	Vector<Int> nIterPerAve;

	// make sure user set avgtime value, else get AveragingTVI error
	if (averaging_.time() && averaging_.timeValue()==0.0) {
		averaging_.setTime(false);
		logWarn(PMS::LOG_ORIGIN_LOAD_CACHE, "Time averaging disabled: value is zero."); 
	}
	// only use scalarAve if other averaging enabled
	bool useScalarAve = averaging_.scalarAve() && (averaging_.time() ||
		averaging_.baseline() || averaging_.antenna() ||  averaging_.spw());
	if (averaging_.scalarAve() && !useScalarAve)
		logWarn(PMS::LOG_ORIGIN_LOAD_CACHE, "Scalar averaging ignored: no other averaging is enabled.");
	averaging_.setScalarAve(useScalarAve);

	if ( averaging_.baseline() || averaging_.antenna() || averaging_.spw() || useScalarAve) {
		// Averaging with PlotMSVBAverager
		// Create visibility iterator vi_p
		setUpVisIter(selection_, calibration_, dataColumn_, loadAxes, loadData);

		// Set nIterPerAve (number of chunks per average)
		bool chunksCounted = countChunks(*vi_p, nIterPerAve, loadAxes, loadData, thread);
		if (chunksCounted) {
			try {
				trapExcessVolume(pendingLoadAxes_);  // check mem req using VolMeter
				deleteVm();
				loadChunks(*vi_p, averaging_, nIterPerAve, loadAxes, loadData, thread, needStokesConverter);
			} catch(const AipsError& log) {
				deleteVm();
				loadError(log.getMesg());
			}
		}	
	} else {
		// Averaging with TransformingVI2 
		try {
			// setUpVisIter also gets the VB shapes and calls trapExcessVolume:
			setUpVisIter(selection_, calibration_, dataColumn_, loadAxes, loadData, false, true, thread);
			loadChunks(*vi_p, loadAxes, loadData, thread, needStokesConverter);
		} catch(const AipsError& log) {
			loadError(log.getMesg());
		}
	}
	// Remember # of VBs per Average
	nVBPerAve_.resize();
	if (nIterPerAve.nelements()>0) {
		nVBPerAve_ = nIterPerAve;
	} else {
		nVBPerAve_.resize(nChunk_);
		nVBPerAve_.set(1);
	}

	completeLoadPageHeaderCache();

	deleteVi(); // close any open tables
}

void MSCache::loadError(String mesg) {
	// catch load error, clear the existing cache, and rethrow
	logLoad(mesg);
	clear();
	deleteVi();  // close any open tables
	deleteAtm();
	stringstream ss;
	ss << mesg;
	throw(AipsError(ss.str()));
}

void MSCache::deleteVi() {
	if (vi_p) delete vi_p;
	vi_p = NULL;
}

void MSCache::deleteVm() {
	if (vm_) delete vm_;
	vm_ = NULL;
}

String MSCache::getDataColumn(vector<PMS::Axis>& loadAxes,
							  vector<PMS::DataColumn>& loadData)
{   // Check data column choice and determine which column to pass to VisIter
	String dataColumn = "NONE";  // default is none - CAS-7506
	std::set<String> dataCols;
	
	// Get datacolumn for visibility & weight axes only
	for (uInt i=0; i<loadAxes.size(); ++i) {
		PMS::Axis thisAxis = loadAxes[i];
		if (PMS::axisIsData(thisAxis) || PMS::axisIsWeight(thisAxis)) {
			// check if requested data column exists in MS
			PMS::DataColumn adjustedCol = checkRequestedDataColumn(loadData[i]);
			if (adjustedCol != loadData[i])
				adjustCurrentAxes(thisAxis, loadData[i], adjustedCol);
			loadData[i] = adjustedCol;
			
			// check if axis/datacol combo valid
			switch (thisAxis) {
				case PMS::AMP:
				case PMS::REAL:
				case PMS::WTxAMP:
				case PMS::SIGMA:
				case PMS::SIGMASP: {
					dataColumn = PMS::dataColumn(loadData[i]);
					break;
				}
				case PMS::PHASE:
				case PMS::IMAG: {
					// These axes not valid with float data
					if (loadData[i] == PMS::FLOAT_DATA) {
						throw(AipsError("Chosen axis not valid for FLOAT_DATA, please use AMP or change Data Column"));
					} else {
						dataColumn = PMS::dataColumn(loadData[i]);
					}
					break;
				}
				case PMS::WTSP: {
					// CAS-7517 wtsp col exists but is empty - plot weight instead
					Table thisTable(filename_);
					const ColumnDescSet cds = thisTable.tableDesc().columnDescSet();
					if (cds.isDefined("WEIGHT_SPECTRUM")) {
						ArrayColumn<Float> weightSpectrum;
						weightSpectrum.attach(thisTable,
							MS::columnName(MS::WEIGHT_SPECTRUM));
						if (!weightSpectrum.hasContent()) {
							logWarn("load_cache", "Plotting WEIGHT column, WEIGHT_SPECTRUM (WTSP) has not been initialized (this can be changed with initweights task)");
							cout << "WARNING: Plotting WEIGHT column, WEIGHT_SPECTRUM (WTSP) has not been initialized (this can be changed with initweights task)" << endl;
						}
					}
					dataColumn = PMS::dataColumn(loadData[i]);
					break;
				}
				case PMS::WT: {
					// CAS-8895 warn user requested mean WEIGHT not channelized WEIGHT_SPECTRUM
					Table thisTable(filename_);
					const ColumnDescSet cds = thisTable.tableDesc().columnDescSet();
					if (cds.isDefined("WEIGHT_SPECTRUM")) {
						ArrayColumn<Float> weightSpectrum;
						weightSpectrum.attach(thisTable,
							MS::columnName(MS::WEIGHT_SPECTRUM));
						if (weightSpectrum.hasContent()) {
							logLoad("Plotting mean WEIGHT column but WEIGHT_SPECTRUM is available. Request 'WtSp' axis to see channelized weights.");
							cout << "INFO: Plotting mean WEIGHT column but WEIGHT_SPECTRUM is available. Request 'WtSp' axis to see channelized weights." << endl;
						}
					}
					dataColumn = PMS::dataColumn(loadData[i]);
				}
				default:
					break;
			}
			dataCols.insert(dataColumn);
		}
	} 

	// loadAxes is only new axes to load; if no datacolumn,
	// try datacolumn of already-loaded axes
	if (dataColumn == "NONE") {
		dataColumn = checkLoadedAxesDatacol();
		// might still be "NONE"!
	}
	// convert to mstransform datacolumn string
	if (dataColumn != "NONE") {
		if (dataCols.size() > 1)
			dataColumn = "ALL";
		else
			dataColumn = normalizeColumnName(dataColumn);
	}
	return dataColumn;
}

PMS::DataColumn MSCache::checkRequestedDataColumn(PMS::DataColumn requestedDataCol) {
	// Check if requested data, scratch, or float cols exist
	PMS::DataColumn datacol = requestedDataCol;
	Table thisTable(filename_);
	const ColumnDescSet cds = thisTable.tableDesc().columnDescSet();
	Bool datacolOk(cds.isDefined("DATA")), 
		corrcolOk(cds.isDefined("CORRECTED_DATA") ||
			(calibration_.useCallib() && !calibration_.calLibrary().empty())),
		modelcolOk(cds.isDefined("MODEL_DATA")),
		floatcolOk(cds.isDefined("FLOAT_DATA"));

	switch (requestedDataCol) {
		case PMS::DATA: {
			// CAS-7482 - for singledish, use FLOAT if no DATA 
			if (!datacolOk && floatcolOk) {
				datacol = PMS::FLOAT_DATA;
				logWarn( "load_cache", "DATA column not present; will use FLOAT_DATA instead.");
			}
			break;
		}
		case PMS::CORRECTED: {
			// requested corrected data but no (real or OTF) corrected column
			if (!corrcolOk) {
				if (datacolOk) { // CAS-5214 - use DATA if no CORRECTED
					datacol = PMS::DATA;
					logWarn( "load_cache", "CORRECTED_DATA column not present and calibration library not set/enabled; using DATA instead.");
				} else if (floatcolOk) { // CAS-7761 - use FLOAT if no CORRECTED
					datacol = PMS::FLOAT_DATA;
					logWarn( "load_cache", "CORRECTED_DATA column not present and calibration library not set/enabled; using FLOAT_DATA instead.");
				}
				break;
			} 
			break;
		}
		case PMS::MODEL: {
			if (floatcolOk && !modelcolOk) { // model not auto-generated for singledish
				throw(AipsError("MODEL_DATA not present; use FLOAT_DATA."));
			}
			break;
		}
		case PMS::CORRMODEL_V:
		case PMS::CORRMODEL_S: {
			if (!corrcolOk && datacolOk) {  // use data residual instead
				datacol = (requestedDataCol==PMS::CORRMODEL_V ? PMS::DATAMODEL_V : PMS::DATAMODEL_S);
				String warn("CORRECTED_DATA column not present and calibration library not set/enabled; using " + PMS::dataColumn(datacol) + " instead of " + PMS::dataColumn(requestedDataCol));
				logWarn("load_cache", warn);
			} else if (!corrcolOk && floatcolOk) {	 // cannot use float residual instead
				  throw(AipsError("CORRECTED_DATA column not present for residuals, use FLOAT_DATA."));
			} else if (!modelcolOk && floatcolOk) { // cannot generate model for singledish
				  throw(AipsError("MODEL_DATA column not present for residuals, use FLOAT_DATA."));
			}
			break;
		}
		case PMS::CORR_DIV_MODEL_V:
		case PMS::CORR_DIV_MODEL_S: {
			if (!corrcolOk && datacolOk) {  // use data residual instead
				datacol = (requestedDataCol==PMS::CORR_DIV_MODEL_V ? PMS::DATA_DIV_MODEL_V : PMS::DATA_DIV_MODEL_S);
				String warn("CORRECTED_DATA column not present and calibration library not set/enabled; using " +
					PMS::dataColumn(datacol) + " instead of " + PMS::dataColumn(requestedDataCol));
				logWarn("load_cache", warn);
			} else if (!corrcolOk && floatcolOk) {	 // cannot use float residual instead
				  throw(AipsError("CORRECTED_DATA column not present for residuals, use FLOAT_DATA."));
			} else if (!modelcolOk && floatcolOk) { // cannot generate model for singledish
				  throw(AipsError("MODEL_DATA column not present for residuals, use FLOAT_DATA."));
			}
			break;
		}
		case PMS::DATAMODEL_V:
		case PMS::DATAMODEL_S: {
			if (!datacolOk && floatcolOk) { // cannot use float residual instead
				 throw(AipsError("Data residuals not valid; use FLOAT_DATA."));
			}
			break;
		}
		case PMS::FLOAT_DATA: {
			// requested float data but no FLOAT column 
			if (!floatcolOk && datacolOk) {
				throw(AipsError("FLOAT_DATA not present; use DATA"));
			}
			break;
		}
		default:
			break;
	}
	return datacol;
}

void MSCache::adjustCurrentAxes(PMS::Axis axis,
		PMS::DataColumn olddata, PMS::DataColumn newdata) {
	// adjust data column for currentX or currentY
	for (uInt a=0; a<currentX_.size(); ++a) {
		if (currentX_[a] == axis && currentXData_[a]==olddata) {
			currentXData_[a] = newdata;
		}
		if (currentY_[a] == axis && currentYData_[a]==olddata) {
			currentYData_[a] = newdata;
		}
	}
}

String MSCache::checkLoadedAxesDatacol() {
	// Check data column of plotted axes
	String loadedCol = "NONE";
	std::set<String> loadedColumns;
	int axesCount = currentX_.size();
	for (int i=0; i<axesCount; ++i) {
		if (PMS::axisIsData(currentX_[i]) || PMS::axisIsWeight(currentX_[i])) 
			loadedColumns.insert(PMS::dataColumn(currentXData_[i]));
		if (PMS::axisIsData(currentY_[i]) || PMS::axisIsWeight(currentY_[i]))
			loadedColumns.insert(PMS::dataColumn(currentYData_[i]));
	}
	if (loadedColumns.size() == 1)
		loadedCol = *loadedColumns.begin();
	else if (loadedColumns.size() > 1)
		loadedCol = "ALL";
	return loadedCol;
}

String MSCache::normalizeColumnName(String plotmscol)
{
	// Convert datacolumn as needed for MSTransformManager
	String colname = plotmscol;
	if ((plotmscol == "corrected-model_vector") || 
		(plotmscol == "corrected-model_scalar") ||
		(plotmscol == "data-model_vector") ||
		(plotmscol == "data-model_scalar") ||
		(plotmscol == "data/model_vector") || 
		(plotmscol == "data/model_scalar") || 
		(plotmscol == "corrected/model_vector") ||
		(plotmscol == "corrected/model_scalar")) {
			colname = "ALL";
	} else if (plotmscol == "float") {
		colname = "FLOAT_DATA";
	} else {   // "data", "corrected", "model"
		colname.upcase();
	}
	return colname;
}

void MSCache::getNamesFromMS(MeasurementSet& ms)
{
	MSColumns msCol(ms);
	antnames_.resize();
	stanames_.resize();
	antstanames_.resize();
	fldnames_.resize();
	intentnames_.resize();
	positions_.resize();

	antnames_	= msCol.antenna().name().getColumn();
	stanames_	= msCol.antenna().station().getColumn();
	antstanames_ = antnames_+String("@")+stanames_;
	positions_   = msCol.antenna().position().getColumn();

	fldnames_	= msCol.field().name().getColumn();

	intentnames_ = msCol.state().obsMode().getColumn();
	mapIntentNamesToIds();  // eliminate duplicate intent names
}

void MSCache::setUpVisIter(PlotMSSelection& selection,
		PlotMSCalibration& calibration,
		String dataColumn, 
		vector<PMS::Axis>& loadAxes,
		vector<PMS::DataColumn>& loadData, 
		Bool interactive,
		Bool estimateMemory,
		ThreadCommunication* thread) {
	/* Create plain or averaging (time or channel) VI with 
		   configuration Record and MSTransformIterator factory */

	// Create configuration:
	// Start with data selection; rename fields with expected keywords
	Record configuration = selection.toRecord();
	configuration.renameField("correlation", configuration.fieldNumber("corr / poln"));
	configuration.renameField("taql", configuration.fieldNumber("msselect"));

	// Add needed fields
	configuration.define("inputms", filename_);
	configuration.define("datacolumn", dataColumn);
	configuration.define("buffermode", true);
	configuration.define("reindex", false);
	configuration.define("interactive", interactive);
	for (auto axis : loadAxes){
		if ( axis == PMS::Axis::RA or axis == PMS::Axis::DEC ) {
			configuration.define("pointingsinterpolation",true);
			break;
		}
	}

	// Add transformations with expected keywords
	configuration.merge(transformations_.toRecord());
	// Use "restfreq" string value
	Double restfreq = configuration.asDouble(configuration.fieldNumber("RestFreq"));
	configuration.removeField(configuration.fieldNumber("RestFreq"));
	configuration.define("restfreq", String::toString(restfreq));
	// Use "outframe" and "veltype"
	configuration.renameField("outframe", configuration.fieldNumber("Frame"));
	configuration.renameField("veltype", configuration.fieldNumber("Veldef"));
	// Use PhaseShiftingTVI
	String phaseCenter = configuration.asString(configuration.fieldNumber("PhaseCenter"));
	configuration.removeField(configuration.fieldNumber("PhaseCenter"));
	if (!phaseCenter.empty()) {
		// Subrecord
		casacore::Record phaseShiftParams;
		phaseShiftParams.define("phasecenter", phaseCenter);

		// Enable phaseshift
		configuration.define("tviphaseshift", true);
		configuration.defineRecord("tviphaseshiftlib", phaseShiftParams);
	}

	// Add calibration library if set
	if (calibration.useCallib()) {
        auto callib_str = calibration.calLibrary();
        if (!callib_str.empty()) {
		    configuration.define("callib", callib_str);
        }
	}

	// Apply averaging
	if (averaging_.time()){
		if (averaging_.scalarAve()) {
			configuration.define("scalaraverage", true);
		} else {
			configuration.define("timeaverage", true);
			String timespanStr = "state";
			if (averaging_.field())
				timespanStr += ",field";
			if (averaging_.scan())
				timespanStr += ",scan";
			configuration.define("timespan", timespanStr);
		}
		configuration.define("timebin", averaging_.timeStr());
	}

	if (averaging_.channel()) {
		int chanBin;
		double chanVal = averaging_.channelValue();
		if (chanVal > INT_MAX) {
			chanBin = INT_MAX;
			logWarn(PMS::LOG_ORIGIN_LOAD_CACHE, "avgchannel value exceeds maximum integer allowed (" + String::toString(INT_MAX) + ")");
		} else if (chanVal == 0.0) {
			chanBin = 1;
			logLoad("Cannot average 0 channels, using 1 instead (no averaging).");
		} else {
			chanBin = static_cast<int>(chanVal);
		}
		configuration.define("chanaverage", true);
		configuration.define("chanbin", chanBin);
	}

	if (averaging_.spw()) {
		configuration.define("spwaverage", true);
	}

	// Filter out MSTransformManager setup messages - set to WARN level then restore
	LogFilter oldFilter(LogMessage::NORMAL);
	if (plotms_) {
		oldFilter = LogFilter(plotms_->getParameters().logPriority());
	}

	// Set up VisibilityIterator using factory
	MSTransformIteratorFactory* factory(nullptr);

	try {
		LogFilter filter(LogMessage::WARN);
		LogSink().globalSink().filter(filter);

		factory = new MSTransformIteratorFactory(configuration);

		if (estimateMemory) {
			if (thread) {
				updateEstimateProgress(thread);
			}

			visBufferShapes_ = factory->getVisBufferStructure();

			// Restore filter and continue
			LogSink().globalSink().filter(oldFilter);

			trapExcessVolume(pendingLoadAxes_);

			Int chunks = visBufferShapes_.size();
			setCache(chunks, loadAxes, loadData);
		} else {
			// Restore filter
			LogSink().globalSink().filter(oldFilter);
			visBufferShapes_.clear();
		}

		vi_p = new vi::VisibilityIterator2(*factory);

		if (factory) delete factory;
	} catch(const AipsError& log) {
		// Restore filter, delete factory, and rethrow
		LogSink().globalSink().filter(oldFilter);

		try {
			if (factory) delete factory;
		} catch(const AipsError& ae) {
		}

		throw(AipsError(log.getMesg()));
	}
}

vi::VisibilityIterator2* MSCache::setUpVisIter(MeasurementSet& selectedMS,
	Vector<Vector<Slice>>& chansel, Vector<Vector<Slice>>& corrsel) {
	// Plain VI2 for chunk counting with baseline/antenna/spw averaging
	// (need to set SortColumns manually)
	Bool combscan(averaging_.scan());
	Bool combfld(averaging_.field());
	Bool combspw(averaging_.spw());

	// Set iterInterval
	Double iterInterval(0.0);
	if (averaging_.time()){
		iterInterval = averaging_.timeValue();
	}
	if (combspw || combfld) iterInterval = DBL_MIN;  // force per-timestamp chunks

	// Create SortColumns
	Int nsortcol(4 + Int(!combscan));  // include room for scan
	Block<Int> columns(nsortcol);
	Int i(0);
	columns[i++]				= MS::ARRAY_ID;
	if (!combscan) columns[i++] = MS::SCAN_NUMBER;  // force scan boundaries
	if (!combfld) columns[i++]  = MS::FIELD_ID;	  // force field boundaries
	if (!combspw) columns[i++]  = MS::DATA_DESC_ID;  // force spw boundaries
	columns[i++]				= MS::TIME;
	if (combfld) columns[i++]   = MS::FIELD_ID;	  // effectively ignore field boundaries
	if (combspw) columns[i++]   = MS::DATA_DESC_ID;  // effectively ignore spw boundaries
	vi::SortColumns sortcol(columns, false);

	vi::VisibilityIterator2* vi2 = new vi::VisibilityIterator2(selectedMS, sortcol, false, 0, iterInterval);
	vi::FrequencySelectionUsingChannels fs;
	setUpFrequencySelectionChannels(fs, chansel);
	fs.addCorrelationSlices(corrsel);
	// Add FrequencySelection to VI
	vi2->setFrequencySelection(fs);
	return vi2;
}

void MSCache::setUpFrequencySelectionChannels(
	vi::FrequencySelectionUsingChannels& fs, Vector<Vector<Slice>>& chansel) {
	/* For the plain VI2 for chunk counting */
	int nSpws = static_cast<int>(chansel.nelements());
	int nChansels;
	for (int spw=0; spw<nSpws; spw++) {
		nChansels = static_cast<int>(chansel[spw].nelements());
		// Add channel selections to FrequencySelection
		for ( int sel=0; sel<nChansels; sel++) {
			fs.add(spw, 
				   chansel[spw][sel].start(), 
				   chansel[spw][sel].length(),
				   chansel[spw][sel].inc());
		}
	}
}

void MSCache::updateEstimateProgress(ThreadCommunication* thread) {
	thread->setStatus("Establishing cache size.  Please wait...");
	thread->setAllowedOperations(false,false,true);
	thread->setProgress(2);
}

bool MSCache::countChunks(vi::VisibilityIterator2& vi,
		Vector<Int>& nIterPerAve,
		vector<PMS::Axis>& loadAxes,
		vector<PMS::DataColumn>& loadData, 
		ThreadCommunication* thread) {
	// Let plotms count the chunks for memory estimation 
	//   when baseline/antenna/spw/scalar averaging
	if (thread != NULL)
		updateEstimateProgress(thread);

	Bool verby(False);
	stringstream ss;

	Bool combscan(averaging_.scan());
	Bool combfld(averaging_.field());
	Bool combspw(averaging_.spw());

	vi::VisBuffer2* vb = vi.getVisBuffer();
	vi.originChunks();
	vi.origin();

	// Keeping time
	Double time1(0.0), avetime1(-1.0);
	Double interval(0.0);
	if (averaging_.time())
		interval = averaging_.timeValue();
	// Keep track of other boundaries
	Int thisscan(-1),lastscan(-1);
	Int thisfld(-1), lastfld(-1);
	Int thisspw(-1),lastspw(-1);
	Int thisddid(-1),lastddid(-1);
	Int thisobsid(-1),lastobsid(-1);
	// Averaging stats
	Int chunk(0), subchunk(0);
	rownr_t maxAveNRows(0);
	nIterPerAve.resize(100);
	nIterPerAve = 0;
	Int nAveInterval(-1);

	for (vi.originChunks(); vi.moreChunks(); vi.nextChunk(), chunk++) {
		subchunk = 0;
		for (vi.origin(); vi.more(); vi.next(), subchunk++) {
			// If a thread is given, check if the user canceled.
			if (thread != NULL) {
				if (thread->wasCanceled()) {
					dataLoaded_ = false;
					userCanceled_ = true;
					return false;
				} else {
					// else users think it's hung...
					if ((chunk % 100) == 0)
						thread->setProgress(chunk/100);
				}
			}

			time1 = vb->time()(0); // first timestamp in this vb
			thisscan = vb->scan()(0);
			thisfld = vb->fieldId()(0);
			thisspw = vb->spectralWindows()(0);
			thisddid = vb->dataDescriptionIds()(0);
			thisobsid = vb->observationId()(0);

			// New ave interval if:
			if ( ((time1-avetime1) > interval) ||		  // exceeded time interval
				 ((time1-avetime1) < 0.0) ||			   // negative timestep
				 (!combscan && (thisscan != lastscan)) ||  // not combing scans, and new scan encountered OR
				 (!combfld && (thisfld != lastfld)) ||	 // not combing fields, and new field encountered OR
				 (!combspw && (thisspw != lastspw)) ||	 // not combing spws, and new spw encountered  OR
				 (thisobsid != lastobsid) ||			   // don't average over obs id
				 (nAveInterval == -1)) {				   // this is the first interval

				if (verby) {
					ss << "--------------------------------\n";
					ss << boolalpha << interval << " "
					   << ((time1 - avetime1)>interval) << " "
					   << ((time1 - avetime1)<0.0) << " "
					   << (!combscan && (thisscan!=lastscan)) << " "
					   << (!combspw && (thisspw!=lastspw)) << " "
					   << (!combfld && (thisfld!=lastfld)) << " "
					   << (!combspw && (thisspw!=lastspw)) << " "
					   << (thisobsid!=lastobsid) << " "
					   << (nAveInterval == -1) << "\n";
				}

				// If we have accumulated enough info, poke the volume meter,
				//  with the _previous_ info, and reset the ave'd row counter
				if (nAveInterval > -1) {
					vm_->add(lastddid, maxAveNRows);
					maxAveNRows = 0;
				}

				nAveInterval++;
				if (verby) ss << "ave = " << nAveInterval << "\n";

				// increase size of nIterPerAve array, if needed
				if (nIterPerAve.nelements() < uInt(nAveInterval+1))
					nIterPerAve.resize(nIterPerAve.nelements()+100, true);
				// initialize next ave interval
				nIterPerAve(nAveInterval) = 0;
				avetime1 = time1;  // first timestamp in this averaging interval
			}

			// Keep track of the maximum # of rows that might get averaged
			maxAveNRows = max(maxAveNRows, vb->nRows());
			// Increment chunk-per-sol count for current solution
			nIterPerAve(nAveInterval)++;

			if (verby) {
				ss << "	 chunk=" << chunk << " subchunk " << subchunk << "\n";
				ss << "		 time=" << vb->time()(0) << " ";
				ss << "arrayId=" << vb->arrayId()(0) << " ";
				ss << "scan" << thisscan << " ";
				ss << "fieldId=" << thisfld << " ";
				ss << "spw=" << thisspw << " ";
				ss << "obsId=" << thisobsid << "\n";
			}

			lastscan = thisscan;
			lastfld  = thisfld;
			lastspw  = thisspw;
			lastddid = thisddid;
			lastobsid = thisobsid;
		}
	}
	// Add in the last iteration
	vm_->add(lastddid,maxAveNRows);

	Int nAve(nAveInterval+1);
	nIterPerAve.resize(nAve, True);
	setCache(nAve, loadAxes, loadData);  // initialize; sets nChunk_

	if (verby) {
		ss << "nIterPerAve = " << nIterPerAve << "\n";
		ss << "Found " << nChunk_ << " chunks." << endl;
		logInfo("count_chunks", ss.str());
	}
	return true;
}

void MSCache::trapExcessVolume(map<PMS::Axis,Bool>& pendingLoadAxes) {
	try {
		String s;
		if (visBufferShapes_.size() > 0) {
			s = vm_->evalVolume(visBufferShapes_, pendingLoadAxes); }
		else {
			Vector<Bool> mask(4, false);
			int dataCount = getDataCount();

			for ( int i = 0; i < dataCount; i++ ){
				Vector<Bool> subMask = netAxesMask( currentX_[i], currentY_[i]);
				mask = mask || subMask;
			}
			s = vm_->evalVolume(pendingLoadAxes, mask);
		}
		logLoad(s);
	} catch(const AipsError& log) {
		// catch detected volume excess, clear the existing cache, and rethrow
		logLoad(log.getMesg());
		deleteVm();
		stringstream ss;
		ss << "Estimated cache exceeds limits." << endl
			<< "Please try using data selection, averaging," << endl
			<< "checking 'Reload' (to clear unneeded cache items),"
			<< "or letting other memory-intensive processes finish.";
		throw(AipsError(ss.str()));
	}
}

void MSCache::updateProgress(ThreadCommunication* thread, Int chunk) {
	double progress;
	// Update progress meter
	if((nChunk_ <= (int)THREAD_SEGMENT || chunk % THREAD_SEGMENT == 0)) {
		thread->setStatus("Loading chunk " + String::toString(chunk) +
				" / " + String::toString(nChunk_) + ".");
		progress = ((double)chunk+1) / nChunk_;
		thread->setProgress((unsigned int)((progress * 100) + 0.5));
	}
}

void MSCache::loadChunks(vi::VisibilityIterator2& vi,
		const vector<PMS::Axis>& loadAxes,
		const vector<PMS::DataColumn>& loadData,
		ThreadCommunication* thread,
		bool needStokesConverter) {

	// permit cancel in progress meter:
	if(thread != NULL)
		thread->setAllowedOperations(false,false,true);
	logLoad("Loading chunks......");

	// Initialize VI and get VB
	vi::VisBuffer2* vb = vi.getVisBuffer();
	vi.originChunks();
	try {
		vi.origin();
	} catch (const ArraySlicerError& err) {
		throw(AipsError("PlotMS averaging error: Data shapes do not conform."));
	} catch (const ArrayConformanceError& err) {
		throw(AipsError("PlotMS averaging error: Data shapes do not conform."));
	}

	nAnt_ = vb->nAntennas();  // needed to set up indexer
	// set frame
	freqFrame_ = transformations_.frame();
	if (freqFrame_ == MFrequency::N_Types) {
		freqFrame_ = static_cast<MFrequency::Types>(vi_p->getReportingFrameOfReference());
	}

	Int chunk(0), lastscan(0), thisscan(0), lastspw(-1), thisspw(0);
	chshapes_.resize(4,nChunk_);
	goodChunk_.resize(nChunk_);
	goodChunk_.set(false);

	// Prepare for loading RA and DEC axes
	using PI_TVI = vi::PointingInterpolationTVI;
	PI_TVI* piTvi = nullptr;
	auto isRaOrDec = [](PMS::Axis a) { return a == PMS::RA || a == PMS::DEC ; };
	auto loadRaOrDec = std::find_if(loadAxes.begin(),loadAxes.end(),isRaOrDec) != loadAxes.end();
	if (loadRaOrDec) {  // setupRaDecIterator
		// Get iterator's implementation, because we need to configure it at run-time
		auto * viImpl = vi.getImpl();
		auto * transformIt = dynamic_cast<MSTransformIterator *>(viImpl);
		if (transformIt != nullptr)
			piTvi = dynamic_cast<PI_TVI *>(transformIt->getInputViIterator());
		if (piTvi == nullptr)
			throw(AipsError("MSCache::loadChunks(): error while preparing for pointings interpolation"));
		piTvi_ = piTvi;
	}


	for(vi.originChunks(); vi.moreChunks(); vi.nextChunk()) {
		for(vi.origin(); vi.more(); vi.next()) {
			if (vb->nRows() <= 0) continue;

			if (chunk >= nChunk_) {  // nChunk_ was just an estimate
				bool increaseChunks(true);
				setCache(chunk+1, loadAxes, loadData, increaseChunks);
				chshapes_.resize(4, nChunk_, true);
				goodChunk_.resize(nChunk_, true);
			}

			// If a thread is given, update its chunk number and progress bar
			if (thread != NULL) updateProgress(thread, chunk);

			// Cache the data shapes
			casacore::IPosition data_shape = vb->flagCube().shape();
			chshapes_(0,chunk) = data_shape(0);
			chshapes_(1,chunk) = data_shape(1);
			chshapes_(2,chunk) = data_shape(2);
			chshapes_(3,chunk) = vb->nAntennas();
			goodChunk_(chunk)  = true;

			// Set up calculator for converted stokes if needed
			casacore::StokesConverter* stokesConverter(nullptr);
			if (needStokesConverter) {
				stokesConverter = getStokesConverter(vb);
				chshapes_(0, chunk) = stokes_out_.size();
			}

			// Load axes
			set<DirectionAxisParams> raDecLoaded;
			for(unsigned int i = 0; i < loadAxes.size(); i++) {
				// If a thread is given, check if the user canceled.
				if(thread != NULL && thread->wasCanceled()) {
					dataLoaded_ = false;
					userCanceled_ = true;
					goodChunk_(chunk) = false; //only partially loaded
					return;
				}
				if (isRaOrDec(loadAxes[i])){  // setupRaDecInterpolator(i,raDecLoaded);
					DirectionAxisParams raDecParams {loadXYFrame_[i],loadXYInterp_[i]};
					auto raDecChunkLoaded = ( raDecLoaded.find(raDecParams) != raDecLoaded.end() );
					if (raDecChunkLoaded) continue;
					raDecLoaded.insert(raDecParams);
					using TviInterp = PI_TVI::Interpolator::InterpMethod;
					auto & interpolator = piTvi->getInterpolator();
					switch (loadXYInterp_[i]){
					case PMS::InterpMethod::NEAREST:
						interpolator.setInterpMethod(TviInterp::NEAREST);
						break;
					case PMS::InterpMethod::CUBIC_SPLINE:
						interpolator.setInterpMethod(TviInterp::CUBIC_SPLINE);
						break;
					default:
						String errMsg("MSCache::loadChunks(): unsupported pointing interpolation method: ");
						errMsg += PMS::interpMethod(loadXYInterp_[i]);
						throw(AipsError(errMsg));
					}
					switch (loadXYFrame_[i]){
					case PMS::CoordSystem::AZELGEO:
						piTvi->setOutputDirectionFrame(MDirection::AZELGEO);
						break;
					case PMS::CoordSystem::ICRS:
						piTvi->setOutputDirectionFrame(MDirection::ICRS);
						break;
					case PMS::CoordSystem::J2000:
						piTvi->setOutputDirectionFrame(MDirection::J2000);
						break;
					case PMS::CoordSystem::B1950:
						piTvi->setOutputDirectionFrame(MDirection::B1950);
						break;
					case PMS::CoordSystem::GALACTIC:
						piTvi->setOutputDirectionFrame(MDirection::GALACTIC);
						break;
					}
					auto & raBlock = raMap_.at(raDecParams);
					loadRa_ = &raBlock;
					auto & decBlock = decMap_.at(raDecParams);
					loadDec_ = &decBlock;
				}

				loadAxis(vb, chunk, loadAxes[i], loadData[i], stokesConverter);

				if (loadAxes[i]==PMS::ATM || loadAxes[i]==PMS::TSKY) {
					thisscan = vb->scan()(0);
					if (thisscan != lastscan) {
						printAtmStats(thisscan);
						lastscan = thisscan;
					}
					thisspw = vb->spectralWindows()(0);
					if (thisspw != lastspw) {
						uInt vectorsize = ( loadAxes[i]==PMS::ATM ?
							(*atm_[chunk]).nelements() :
							(*tsky_[chunk]).nelements());
						if (vectorsize==1) {
							logWarn("load_cache", "Setting " +
							  PMS::axis(loadAxes[i]) + " for spw " +
							  String::toString(thisspw) +
							  " to zero because it has only one channel.");
						}
						lastspw = thisspw;
					}
				}
			}
			chunk++;
            if (needStokesConverter) {
				delete stokesConverter;
			}
		}
	}
	// Report averaged channels per spw in log
	// (should match MSTransformManager output in console)
	map<Int,Int>::iterator it;
	for (it=chansPerSpw_.begin(); it!=chansPerSpw_.end(); ++it) {
		logLoad("SPW " + String::toString(it->first) + ": number of channels averaged = " + String::toString(it->second));
	}
}

void MSCache::loadChunks(vi::VisibilityIterator2& vi,
		const PlotMSAveraging& averaging,
		const Vector<Int>& nIterPerAve,
		const vector<PMS::Axis>& loadAxes,
		const vector<PMS::DataColumn>& loadData,
		ThreadCommunication* thread,
		bool needStokesConverter) {

	// permit cancel in progress meter:
	if(thread != NULL)
		thread->setAllowedOperations(false,false,true);
	logLoad("Loading chunks with averaging.....");

	Bool verby(false);

	vi::VisBuffer2* vb = vi.getVisBuffer();
	vi.originChunks();
	vi.origin();
	nAnt_ = vb->nAntennas();  // needed to set up indexer
	// set frame
	freqFrame_ = transformations_.frame();
	if (freqFrame_ == MFrequency::N_Types) {
		freqFrame_ = static_cast<MFrequency::Types>(vi_p->getReportingFrameOfReference());
	}

	chshapes_.resize(4, nChunk_);
	goodChunk_.resize(nChunk_);
	goodChunk_.set(false);
	Int nAnts;
	vi::VisBuffer2* vbToUse(NULL);
	casacore::Int lastscan(0), thisscan(0);  // for printing atm stats
	casacore::Int lastspw(-1), thisspw(0);  // for printing ATM/TSKY warning 

	for (Int chunk=0; chunk<nChunk_; ++chunk) {
		if (chunk >= nChunk_) {  // nChunk_ was just an estimate!
			bool increaseChunks(true);
			setCache(chunk, loadAxes, loadData, increaseChunks);
			chshapes_.resize(4, nChunk_, true);
			goodChunk_.resize(nChunk_, true);
		}

		// If a thread is given, update it.
		if(thread != NULL)
			updateProgress(thread, chunk);

		// Set up VB averager
		nAnts = vb->nAntennas();
		PlotMSVBAverager pmsvba(nAnts);
		pmsvba.setBlnAveraging(averaging.baseline());
		pmsvba.setAntAveraging(averaging.antenna());
		pmsvba.setScalarAve(averaging.scalarAve());
		// Sort out which data to read
		discernData(loadAxes,loadData,pmsvba);

		stringstream ss;
		if (verby) ss << "Chunk " << chunk << " ----------------------------------\n";

		Int iter=0;
		while (iter < nIterPerAve(chunk)) {
			if (verby) {
				ss << "chunk=" << chunk << " iter=" << iter << " nIterPerAve =" << nIterPerAve(chunk) << " ";
				ss << "scan=" << vb->scan()(0) << " ";
				ss << "fieldId=" << vb->fieldId()(0) << " ";
				ss << "spw=" << vb->spectralWindows()(0) << " ";
				ss << "amp=" << vb->visCube().shape() << " ";
				ss << "freq=" << vb->getFrequencies(0, freqFrame_).shape() << " ";
			}
			// Accumulate into the averager
			pmsvba.accumulate(*vb);
			
			// Advance to next VB unless you are going to finalize
			if (iter+1 < nIterPerAve(chunk)) {
				if (verby) ss << " next VB "; 
				vi.next(); 
				if (!vi.more() && vi.moreChunks()) { 
					 // go to first vb in next chunk 
					 if (verby) ss << "  stepping VI"; 
					 vi.nextChunk(); 
					 vi.origin(); 
				}
			}
			if (verby) ss << "\n";
			++iter;
		}

		if (verby) ss << "Finalize average\n";
		// Finalize the averaging
		pmsvba.finalizeAverage();
		// The averaged VisBuffer
		vi::VisBuffer2& avb(pmsvba.aveVisBuff());
		// Only if the average yielded some data:
		if (avb.nRows() > 0) {
			// Cache the data shapes
			chshapes_(0,chunk) = avb.nCorrelations();
			chshapes_(1,chunk) = avb.nChannels();
			chshapes_(2,chunk) = avb.nRows();
			chshapes_(3,chunk) = nAnts;
			goodChunk_(chunk)  = true;

			// Set up calculator for converted stokes if needed
			casacore::StokesConverter* stokesConverter(nullptr);
			if (needStokesConverter) {
				stokesConverter = getStokesConverter(vb);
			}

			for(unsigned int i = 0; i < loadAxes.size(); i++) {
				// If a thread is given, check if the user canceled.
				if(thread != NULL && thread->wasCanceled()) {
					dataLoaded_ = false;
					userCanceled_ = true;
					goodChunk_(chunk)  = false;
					return;
				}
				if (useAveragedVisBuffer(loadAxes[i])) {
					vbToUse = &avb;
				} else {
					vbToUse = vb;
				}

				loadAxis(vbToUse, chunk, loadAxes[i], loadData[i], stokesConverter);

				if (loadAxes[i]==PMS::ATM || loadAxes[i]==PMS::TSKY) {
					thisscan = vbToUse->scan()(0);
					if (thisscan != lastscan) {
						printAtmStats(thisscan);
						lastscan = thisscan;
					}
					thisspw = vbToUse->spectralWindows()(0);
					if (thisspw != lastspw) {
						uInt vectorsize = ( loadAxes[i]==PMS::ATM ?
							(*atm_[chunk]).nelements() :
							(*tsky_[chunk]).nelements());
						if (vectorsize==1) {
							logWarn("load_cache", "Setting " + 
							  PMS::axis(loadAxes[i]) + " for spw " +
							  String::toString(thisspw) +
							  " to zero because it has only one channel.");
						}
						lastspw = thisspw;
					} 
				}
			}
		} else {
			// no points in this chunk
			goodChunk_(chunk) = false;
			chshapes_.column(chunk) = 0;
		}
		// Now advance to next chunk
		if (verby) ss << " next VB "; 
		vi.next();
		if (!vi.more() && vi.moreChunks()) { 
			 // go to first vb in next chunk 
			 if (verby) ss << "  stepping VI"; 
			 vi.nextChunk(); 
			 vi.origin(); 
		}
		if(verby) {
			ss << "\n";
			logLoad(ss.str());
		}
	}
	//cout << boolalpha << "goodChunk_ = " << goodChunk_ << endl;
}

casacore::StokesConverter* MSCache::getStokesConverter(vi::VisBuffer2* vb) {
	// Set up stokes converter with defined correlation types in this VisBuffer
	auto definedCorrTypes = vb->getCorrelationTypesDefined();
    size_t ncorr = definedCorrTypes.size();
	Vector<int> stokes_in(ncorr);

	for (size_t i = 0; i < ncorr; ++i) {
		stokes_in[i] = definedCorrTypes[i];
	}
	return new StokesConverter(stokes_out_, stokes_in);
}

bool MSCache::useAveragedVisBuffer(PMS::Axis axis) {
	// Some axes should be obtained from the VB2 provided by the VI2
	// rather than the averaged VB2, which is not attached
	bool useAvg(true);
	switch(axis) {
	case PMS::CHANNEL:
	case PMS::FREQUENCY:
	case PMS::VELOCITY:
	case PMS::UVDIST_L:
	case PMS::UWAVE:
	case PMS::VWAVE:
	case PMS::WWAVE:
	case PMS::AZ0:
	case PMS::EL0:
	case PMS::HA0:
	case PMS::PA0:
	case PMS::ANTENNA:
	case PMS::AZIMUTH:
	case PMS::RA:
	case PMS::DEC:
	case PMS::ELEVATION:
	case PMS::PARANG:
	case PMS::ROW:
	case PMS::ATM:
	case PMS::TSKY:
	case PMS::IMAGESB: {
		useAvg = false;
		break;
	}
	default:
		break;
	}
	return useAvg;
}

void MSCache::forceVBread(vi::VisBuffer2* vb,
		const vector<PMS::Axis>& loadAxes,
		const vector<PMS::DataColumn>& loadData) {

	// pre-load requisite pieces of VisBuffer for averaging
	for(unsigned int i = 0; i < loadAxes.size(); i++) {
		switch (loadAxes[i]) {
		case PMS::AMP:
		case PMS::PHASE:
		case PMS::REAL:
		case PMS::IMAG: {
			switch(loadData[i]) {
			case PMS::DATA: {
				vb->visCube();
				break;
			}
			case PMS::MODEL: {
				vb->visCubeModel();
				break;
			}
			case PMS::CORRECTED: {
				vb->visCubeCorrected();
				break;
			}
			case PMS::CORRMODEL_V:
			case PMS::CORRMODEL_S:
			case PMS::CORR_DIV_MODEL_V:
			case PMS::CORR_DIV_MODEL_S: {
				vb->visCubeCorrected();
				vb->visCubeModel();
				break;
			}
			case PMS::DATAMODEL_V:
			case PMS::DATAMODEL_S: {
				vb->visCube();
				vb->visCubeModel();
				break;
			}
			case PMS::DATA_DIV_MODEL_V:
			case PMS::DATA_DIV_MODEL_S: {
				vb->visCube();
				vb->visCubeModel();
				break;
			}
			case PMS::FLOAT_DATA: {
				vb->visCubeFloat();
				break;
			}
			default:
				break;
			}
			break;
		}
		default:
			break;
		}
	}

	// Always need flags
	vb->flagRow();
	vb->flagCube();

}

void MSCache::discernData(const vector<PMS::Axis>& loadAxes,
	const vector<PMS::DataColumn>& loadData,
	PlotMSVBAverager& vba) {

	// Turn off
	vba.setNoData();

	// Tell the averager which data column to read
	for(unsigned int i = 0; i < loadAxes.size(); i++) {
		switch (loadAxes[i]) {
		case PMS::AMP:
		case PMS::PHASE:
		case PMS::REAL:
		case PMS::IMAG: {
			switch(loadData[i]) {
			case PMS::DATA: {
				vba.setDoVC();
				break;
			}
			case PMS::MODEL: {
				vba.setDoMVC();
				break;
			}
			case PMS::CORRECTED: {
				vba.setDoCVC();
				break;
			}
			case PMS::CORRMODEL_V:
			case PMS::CORRMODEL_S:
			case PMS::CORR_DIV_MODEL_V:
			case PMS::CORR_DIV_MODEL_S: {
				vba.setDoCVC();
				vba.setDoMVC();
				break;
			}
			case PMS::DATAMODEL_V:
			case PMS::DATAMODEL_S:
			case PMS::DATA_DIV_MODEL_V: 
			case PMS::DATA_DIV_MODEL_S: {
				vba.setDoVC();
				vba.setDoMVC();
				break;
			}
			case PMS::FLOAT_DATA:
				vba.setDoFC();
				break;
			default:
				break;
			}
			break;
		}
		case PMS::UVDIST:
		case PMS::UVDIST_L:
		case PMS::U:
		case PMS::V:
		case PMS::W:
		case PMS::UWAVE:
		case PMS::VWAVE:
		case PMS::WWAVE: {
			vba.setDoUVW();
			break;
		}
		default:
			break;
		}
	}

}

void MSCache::loadAxis(vi::VisBuffer2* vb, Int vbnum, PMS::Axis axis,
		PMS::DataColumn data, StokesConverter* stokesConverter) {
	switch(axis) {

	case PMS::SCAN: { // assumes scan unique in VB
		scan_(vbnum) = vb->scan()(0);
		break;
	}
	case PMS::FIELD: { // assumes field unique in VB
		field_(vbnum) = vb->fieldId()(0);
		break;
	}
	case PMS::TIME: { // assumes time unique in VB
		time_(vbnum) = vb->time()(0);
		break;
	}
	case PMS::TIME_INTERVAL: { // assumes timeInterval unique in VB
		timeIntr_(vbnum) = vb->timeInterval()(0);
		break;
	}
	case PMS::SPW: {
		spw_(vbnum) = vb->spectralWindows()(0);
		break;
	}
	case PMS::CHANNEL: {
		Vector<Int> chans = vb->getChannelNumbers(0);
		*chan_[vbnum] = chans;
		if (averaging_.channel()) {
			// Save which channels are being averaged, for Locate
			Int numBins = chans.size();
			Int numChans = vb->getChannelNumbersSelected(0).size();
			Array<Int> chansPerBin(IPosition(2, numChans, numBins));
			// when there is only one channel in the spw selected chans is empty array
			if (numChans==0 && numBins==1) {
				numChans = 1;
				chansPerBin.resize(IPosition(2, 1, 1));
				chansPerBin[0] = chans;
			} else {
				for (Int bin=0; bin < numBins; ++bin) {
					chansPerBin[bin] = vb->getChannelNumbersSelected(bin);
				}
			}
			*chansPerBin_[vbnum] = chansPerBin;
			chansPerSpw_[vb->spectralWindows()(0)] = numChans;
		}
		break;
	}
	case PMS::FREQUENCY: {
		// Convert freq to desired frame
		  *freq_[vbnum] = vb->getFrequencies(0, freqFrame_);
		(*freq_[vbnum]) /= 1.0e9; // in GHz
		break;
	}
	case PMS::VELOCITY: {
		*vel_[vbnum] = calcVelocity(vb); 
		break;
	}
	case PMS::CORR: {
		if (stokesConverter) {
			*corr_[vbnum] = stokes_out_;
		} else {
			Vector<Stokes::StokesTypes> corrTypes = vb->getCorrelationTypesSelected();
			Vector<Int> corrTypesInt;
			corrTypesInt.resize(corrTypes.size());
			for (uInt i = 0; i < corrTypes.size(); ++i) {
				corrTypesInt[i] = static_cast<Int>(corrTypes[i]);
			}
			*corr_[vbnum] = corrTypesInt;
		}
		break;
	}
	case PMS::ANTENNA1: {
		*antenna1_[vbnum] = vb->antenna1();
		break;
	}
	case PMS::ANTENNA2: {
		*antenna2_[vbnum] = vb->antenna2();
		break;
	}
	case PMS::BASELINE: {
		Vector<Int> a1(vb->antenna1());
		Vector<Int> a2(vb->antenna2());
		baseline_[vbnum]->resize(vb->nRows());
		Vector<Int> bl(*baseline_[vbnum]);
		for (rownr_t irow = 0; irow < vb->nRows(); ++irow) {
			if (a1(irow)<0) a1(irow)=chshapes_(3,0);
			if (a2(irow)<0) a2(irow)=chshapes_(3,0);
			bl(irow)=(chshapes_(3,0)+1)*a1(irow) - (a1(irow) * (a1(irow) + 1)) / 2 + a2(irow);
		}
		break;
	}
	case PMS::UVDIST: {
		Array<Double> u(vb->uvw().row(0));
		Array<Double> v(vb->uvw().row(1));
		*uvdist_[vbnum] = sqrt(u*u+v*v);
		break;
	}
	case PMS::U: {
		*u_[vbnum] = vb->uvw().row(0);
		break;
	}
	case PMS::V: {
		*v_[vbnum] = vb->uvw().row(1);
		break;
	}
	case PMS::W: {
		*w_[vbnum] = vb->uvw().row(2);
		break;
	}
	case PMS::UVDIST_L: {
		Array<Double> u(vb->uvw().row(0));
		Array<Double> v(vb->uvw().row(1));
		Vector<Double> uvdistM = sqrt(u*u + v*v);
		uvdistM /=C::c;
		uvdistL_[vbnum]->resize(vb->nChannels(), vb->nRows());
		Vector<Double> uvrow;
		for (rownr_t irow = 0; irow < vb->nRows(); ++irow) {
			uvrow.reference(uvdistL_[vbnum]->column(irow));
			uvrow.set(uvdistM(irow));
			uvrow *= vb->getFrequencies(irow, freqFrame_);
		}
		break;
	}
	case PMS::UWAVE: {
		Vector<Double> uM(vb->uvw().row(0));
		uM/=C::c;
		uwave_[vbnum]->resize(vb->nChannels(), vb->nRows());
		Vector<Double> urow;
		for (rownr_t irow = 0; irow < vb->nRows(); ++irow) {
			urow.reference(uwave_[vbnum]->column(irow));
			urow.set(uM(irow));
			urow *= vb->getFrequencies(irow, freqFrame_);
		}
		break;
	}
	case PMS::VWAVE: {
		Vector<Double> vM(vb->uvw().row(1));
		vM/=C::c;
		vwave_[vbnum]->resize(vb->nChannels(), vb->nRows());
		Vector<Double> vrow;
		for (rownr_t irow = 0; irow < vb->nRows(); ++irow) {
			vrow.reference(vwave_[vbnum]->column(irow));
			vrow.set(vM(irow));
			vrow *= vb->getFrequencies(irow, freqFrame_);
		}
		break;
	}
	case PMS::WWAVE: {
		Vector<Double> wM(vb->uvw().row(2));
		wM/=C::c;
		wwave_[vbnum]->resize(vb->nChannels(), vb->nRows());
		Vector<Double> wrow;
		for (rownr_t irow = 0; irow < vb->nRows(); ++irow) {
			wrow.reference(wwave_[vbnum]->column(irow));
			wrow.set(wM(irow));
			wrow *= vb->getFrequencies(irow, freqFrame_);
		}
		break;
	}
	case PMS::AMP:
	case PMS::PHASE:
	case PMS::REAL:
	case PMS::IMAG: {
		// Get data cube(s) for data column, converting stokes if needed.
		// Then set axis cube.
		casacore::Cube<casacore::Complex> vis_data;
		casacore::Cube<casacore::Complex> model_data;
		casacore::Cube<casacore::Complex> corrected_data;
		bool is_scalar_avg = averaging_.scalarAve();

		switch(data) {
			case PMS::DATA: {
				auto vis_cube = vb->visCube();
				if (stokesConverter) {
					stokesConverter->convert(vis_data, vis_cube);
				} else {
					vis_data.reference(vis_cube);
				}
				if (axis == PMS::AMP) {
					MeasurementSet ms(filename_);
					if (ms.isColumn(MS::FLOAT_DATA)) {
						*amp_[vbnum] = real(vis_data);
					} else {
						*amp_[vbnum] = (is_scalar_avg ? real(vis_data) : amplitude(vis_data));
					}
				} else if (axis == PMS::PHASE) {
					*pha_[vbnum] = (is_scalar_avg ? imag(vis_data) : phase(vis_data)) * 180.0 / C::pi;
				} else if (axis == PMS::REAL) {
					*real_[vbnum] = real(vis_data);
				} else { // PMS::IMAG
					*imag_[vbnum] = imag(vis_data);
				}
				break;
			}
			case PMS::MODEL: {
				auto vis_cube_model = vb->visCubeModel();
				if (stokesConverter) {
					stokesConverter->convert(model_data, vis_cube_model);
				} else {
					model_data.reference(vis_cube_model);
				}
				if (axis == PMS::AMP) {
					*ampModel_[vbnum] = (is_scalar_avg ? real(model_data) : amplitude(model_data));
				} else if (axis == PMS::PHASE) {
					*phaModel_[vbnum] = (is_scalar_avg ? imag(model_data) : phase(model_data)) * 180.0 / C::pi;
				} else if (axis == PMS::REAL) {
					*realModel_[vbnum] = real(model_data);
				} else { // PMS::IMAG
					*imagModel_[vbnum] = imag(model_data);
				}
				break;
			}
			case PMS::CORRECTED: {
				auto vis_cube_corrected = vb->visCubeCorrected();
				if (stokesConverter) {
					stokesConverter->convert(corrected_data, vis_cube_corrected);
				} else {
					corrected_data.reference(vis_cube_corrected);
				}
				if (axis == PMS::AMP) {
					*ampCorr_[vbnum] = (is_scalar_avg ? real(corrected_data) : amplitude(corrected_data));
				} else if (axis == PMS::PHASE) {
					*phaCorr_[vbnum] = (is_scalar_avg ? imag(corrected_data) : phase(corrected_data)) * 180.0 / C::pi;
				} else if (axis == PMS::REAL) {
					*realCorr_[vbnum] = real(corrected_data);
				} else { // PMS::IMAG
					*imagCorr_[vbnum] = imag(corrected_data);
				}
				break;
			}
			case PMS::CORRMODEL_V:
			case PMS::CORRMODEL_S:
			case PMS::CORR_DIV_MODEL_V:
			case PMS::CORR_DIV_MODEL_S: {
				auto vis_cube_corrected = vb->visCubeCorrected();
				auto vis_cube_model = vb->visCubeModel();
				if (stokesConverter) {
					stokesConverter->convert(corrected_data, vis_cube_corrected);
					stokesConverter->convert(model_data, vis_cube_model);
				} else {
					corrected_data.reference(vis_cube_corrected);
					model_data.reference(vis_cube_model);
				}
				if (data == PMS::CORRMODEL_V) {
					auto diff_data = corrected_data - model_data;
					if (axis == PMS::AMP) {
						*ampCorrModel_[vbnum] = (is_scalar_avg ? real(diff_data) : amplitude(diff_data));
					} else if (axis == PMS::PHASE) {
						*phaCorrModel_[vbnum] = (is_scalar_avg ? imag(diff_data) : phase(diff_data)) * 180.0 / C::pi;
					} else if (axis == PMS::REAL) {
						*realCorrModel_[vbnum] = real(diff_data);
					} else { // PMS::IMAG
						*imagCorrModel_[vbnum] = imag(diff_data);
					}
				} else if (data == PMS::CORRMODEL_S) {
					if (axis == PMS::AMP) {
						*ampCorrModelS_[vbnum] = (is_scalar_avg ? real(corrected_data) - real(model_data) :
							amplitude(corrected_data) - amplitude(model_data));
					} else if (axis == PMS::PHASE) {
						*phaCorrModelS_[vbnum] = (is_scalar_avg ? imag(corrected_data) - imag(model_data) :
							phase(corrected_data) - phase(model_data)) * 180.0 / C::pi;
					} else if (axis == PMS::REAL) {
						*realCorrModelS_[vbnum] = real(corrected_data) - real(model_data);
					} else { // PMS::IMAG
						*imagCorrModelS_[vbnum] = imag(corrected_data) - imag(model_data);
					}
				} else if (data == PMS::CORR_DIV_MODEL_V) {
					auto div_data = corrected_data / model_data;
					if (axis == PMS::AMP) {
						*ampCorrDivModel_[vbnum] = (is_scalar_avg ? real(div_data) : amplitude(div_data));
					} else if (axis == PMS::PHASE) {
						*phaCorrDivModel_[vbnum] = (is_scalar_avg ? imag(div_data) : phase(div_data)) * 180.0 / C::pi;
					} else if (axis == PMS::REAL) {
						*realCorrDivModel_[vbnum] = real(div_data);
					} else { // PMS::IMAG
						*imagCorrDivModel_[vbnum] = imag(div_data);
					}
				} else if (data == PMS::CORR_DIV_MODEL_S) {
					if (axis == PMS::AMP) {
						*ampCorrDivModelS_[vbnum] = (is_scalar_avg ? real(corrected_data) / real(model_data) :
							amplitude(corrected_data) / amplitude(model_data));
					} else if (axis == PMS::PHASE) {
						*phaCorrDivModelS_[vbnum] = (is_scalar_avg ? imag(corrected_data) / imag(model_data) :
							phase(corrected_data) / amplitude(model_data)) * 180.0 / C::pi;
					} else if (axis == PMS::REAL) {
						*realCorrDivModelS_[vbnum] = real(corrected_data) - real(model_data);
					} else { // PMS::IMAG
						*imagCorrDivModelS_[vbnum] = imag(corrected_data) - imag(model_data);
					}
				}
				break;
			}
			case PMS::DATAMODEL_V:
			case PMS::DATAMODEL_S:
			case PMS::DATA_DIV_MODEL_V:
			case PMS::DATA_DIV_MODEL_S: {
				auto vis_cube = vb->visCube();
				auto vis_cube_model = vb->visCubeModel();
				if (stokesConverter) {
					stokesConverter->convert(vis_data, vis_cube);
					stokesConverter->convert(model_data, vis_cube_model);
				} else {
					vis_data.reference(vis_cube);
					model_data.reference(vis_cube_model);
				}
				if (data == PMS::DATAMODEL_V) {
					auto diff_data = vis_data - model_data;
					if (axis == PMS::AMP) {
						*ampDataModel_[vbnum] = (is_scalar_avg ? real(diff_data) : amplitude(diff_data));
					} else if (axis == PMS::PHASE) {
						*phaDataModel_[vbnum] = (is_scalar_avg ? imag(diff_data) : phase(diff_data)) * 180.0 / C::pi;
					} else if (axis == PMS::REAL) {
						*realDataModel_[vbnum] = real(diff_data);
					} else { // PMS::IMAG
						*imagDataModel_[vbnum] = imag(diff_data);
					}
				} else if (data == PMS::DATAMODEL_S) {
					if (axis == PMS::AMP) {
						*ampDataModelS_[vbnum] = (is_scalar_avg ? real(vis_data) - real(model_data) :
							amplitude(vis_data) - amplitude(model_data));
					} else if (axis == PMS::PHASE) {
						*phaDataModelS_[vbnum] = (is_scalar_avg ? imag(vis_data) - imag(model_data) :
							phase(vis_data) - phase(model_data)) * 180.0 / C::pi;
					} else if (axis == PMS::REAL) {
						*realDataModelS_[vbnum] = real(vis_data) - real(model_data);
					} else { // PMS::IMAG
						*imagDataModelS_[vbnum] = imag(vis_data) - imag(model_data);
					}
				} else if (data == PMS::DATA_DIV_MODEL_V) {
					auto div_data = vis_data / model_data;
					if (axis == PMS::AMP) {
						*ampDataDivModel_[vbnum] = (is_scalar_avg ? real(div_data) : amplitude(div_data));
					} else if (axis == PMS::PHASE) {
						*phaDataDivModel_[vbnum] = (is_scalar_avg ? imag(div_data) : phase(div_data)) * 180.0 / C::pi;
					} else if (axis == PMS::REAL) {
						*realDataDivModel_[vbnum] = real(div_data);
					} else { // PMS::IMAG
						*imagDataDivModel_[vbnum] = imag(div_data);
					}
				} else if (data == PMS::DATA_DIV_MODEL_S) {
					if (axis == PMS::AMP) {
						*ampDataDivModelS_[vbnum] = (is_scalar_avg ? real(vis_data) / real(model_data) :
							amplitude(vis_data) / amplitude(model_data));
					} else if (axis == PMS::PHASE) {
						*phaDataDivModelS_[vbnum] = (is_scalar_avg ? imag(vis_data) / imag(model_data) :
							phase(vis_data) / amplitude(model_data)) * 180.0 / C::pi;
					} else if (axis == PMS::REAL) {
						*realDataDivModelS_[vbnum] = real(vis_data) - real(model_data);
					} else { // PMS::IMAG
						*imagDataDivModelS_[vbnum] = imag(vis_data) - imag(model_data);
					}
				}
				break;
			}
			case PMS::FLOAT_DATA: {
				auto float_cube = vb->visCubeFloat();
				if (stokesConverter) {
					casacore::Array<casacore::Float> float_data;
					stokesConverter->convert(float_data, float_cube);
					*ampFloat_[vbnum] = float_data;
				} else {
					*ampFloat_[vbnum] = float_cube;
				}
				break;
			}
		}
		break;
	}
	case PMS::FLAG: {
		auto flag_cube = vb->flagCube();
		casacore::Array<casacore::Bool> flag_data;
		if (stokesConverter) {
			stokesConverter->convert(flag_data, flag_cube);
			*flag_[vbnum] = flag_data;
		} else {
			*flag_[vbnum] = flag_cube;
		}
		break;
	}
	case PMS::FLAG_ROW: {
		*flagrow_[vbnum] = vb->flagRow();
		break;
	}
	case PMS::WT: {
		auto weight = vb->weight();
		if (stokesConverter) {
			casacore::Array<casacore::Float> weight_data;
			stokesConverter->convert(weight_data, weight);
			*wt_[vbnum] = weight_data;
		} else {
			*wt_[vbnum] = weight;
		}
		break;
	}
	case PMS::WTSP: {
		if (vb->weightSpectrum().nelements() > 0) {
			auto weightSpectrum = vb->weightSpectrum();
			if (stokesConverter) {
				casacore::Array<casacore::Float> weightSpectrum_data;
				stokesConverter->convert(weightSpectrum_data, weightSpectrum);
				*wtsp_[vbnum] = weightSpectrum_data;
			} else {
				*wtsp_[vbnum] = weightSpectrum;
			}
		} else {
			throw(AipsError("This MS does not have a valid WEIGHT_SPECTRUM column."));
		}
		break;
	}
	case PMS::WTxAMP: {
		uInt nchannels = vb->nChannels();
		auto weight = vb->weight();
		casacore::Array<casacore::Float> weight_data;
		if (stokesConverter) {
			stokesConverter->convert(weight_data, weight);
		} else {
			weight_data.reference(weight);
		}
		casacore::Cube<Float> wtA;
		switch(data) {
			case PMS::DATA: {
				auto vis_cube = vb->visCube();
				if (stokesConverter) {
					casacore::Cube<casacore::Complex> vis_data;
					stokesConverter->convert(vis_data, vis_cube);
					*wtxamp_[vbnum] = amplitude(vis_data);
				} else {
					*wtxamp_[vbnum] = amplitude(vis_cube);
				}
				wtA.reference(*wtxamp_[vbnum]);
				break;
			}
			case PMS::MODEL: {
				auto model_cube = vb->visCubeModel();
				if (stokesConverter) {
					casacore::Cube<casacore::Complex> model_data;
					stokesConverter->convert(model_data, model_cube);
					*wtxampModel_[vbnum] = amplitude(model_data);
				} else {
					*wtxampModel_[vbnum] = amplitude(model_cube);
				}
				wtA.reference(*wtxampModel_[vbnum]);
				break;
			}
			case PMS::CORRECTED: {
				auto corrected_cube = vb->visCubeCorrected();
				if (stokesConverter) {
					casacore::Cube<casacore::Complex> corrected_data;
					stokesConverter->convert(corrected_data, corrected_cube);
					*wtxampCorr_[vbnum] = amplitude(corrected_data);
				} else {
					*wtxampCorr_[vbnum] = amplitude(corrected_cube);
				}
				wtA.reference(*wtxampCorr_[vbnum]);
				break;
			}
			case PMS::CORRMODEL_V: {
				auto corrected_cube = vb->visCubeCorrected();
				auto model_cube = vb->visCubeModel();
				if (stokesConverter) {
					casacore::Cube<casacore::Complex> corrected_data, model_data;
					stokesConverter->convert(corrected_data, corrected_cube);
					stokesConverter->convert(model_data, model_cube);
					*wtxampCorrModel_[vbnum] = amplitude(corrected_data - model_data);
				} else {
					*wtxampCorrModel_[vbnum] = amplitude(corrected_cube - model_cube);
				}
				wtA.reference(*wtxampCorrModel_[vbnum]);
				break;
			}
			case PMS::CORRMODEL_S: {
				auto corrected_cube = vb->visCubeCorrected();
				auto model_cube = vb->visCubeModel();
				if (stokesConverter) {
					casacore::Cube<casacore::Complex> corrected_data, model_data;
					stokesConverter->convert(corrected_data, corrected_cube);
					stokesConverter->convert(model_data, model_cube);
					*wtxampCorrModelS_[vbnum] = amplitude(corrected_data) - amplitude(model_data);
				} else {
					*wtxampCorrModelS_[vbnum] = amplitude(corrected_cube) - amplitude(model_cube);
				}
				wtA.reference(*wtxampCorrModelS_[vbnum]);
				break;
			}
			case PMS::DATAMODEL_V: {
				auto vis_cube = vb->visCube();
				auto model_cube = vb->visCubeModel();
				if (stokesConverter) {
					casacore::Cube<casacore::Complex> vis_data, model_data;
					stokesConverter->convert(vis_data, vis_cube);
					stokesConverter->convert(model_data, model_cube);
					*wtxampDataModel_[vbnum] = amplitude(vis_data - model_data);
				} else {
					*wtxampDataModel_[vbnum] = amplitude(vis_cube - model_cube);
				}
				wtA.reference(*wtxampDataModel_[vbnum]);
				break;
			}
			case PMS::DATAMODEL_S: {
				auto vis_cube = vb->visCube();
				auto model_cube = vb->visCubeModel();
				if (stokesConverter) {
					casacore::Cube<casacore::Complex> vis_data, model_data;
					stokesConverter->convert(vis_data, vis_cube);
					stokesConverter->convert(model_data, model_cube);
					*wtxampDataModelS_[vbnum] = amplitude(vis_data) - amplitude(model_data);
				} else {
					*wtxampDataModelS_[vbnum] = amplitude(vis_cube) - amplitude(model_cube);
				}
				wtA.reference(*wtxampDataModelS_[vbnum]);
				break;
			}
			case PMS::DATA_DIV_MODEL_V: {
				auto vis_cube = vb->visCube();
				auto model_cube = vb->visCubeModel();
				if (stokesConverter) {
					casacore::Cube<casacore::Complex> vis_data, model_data;
					stokesConverter->convert(vis_data, vis_cube);
					stokesConverter->convert(model_data, model_cube);
					*wtxampDataDivModel_[vbnum] = amplitude(vis_data / model_data);
				} else {
					*wtxampDataDivModel_[vbnum] = amplitude(vis_cube / model_cube);
				}
				wtA.reference(*wtxampDataDivModel_[vbnum]);
				break;
			}
			case PMS::DATA_DIV_MODEL_S: {
				auto vis_cube = vb->visCube();
				auto model_cube = vb->visCubeModel();
				if (stokesConverter) {
					casacore::Cube<casacore::Complex> vis_data, model_data;
					stokesConverter->convert(vis_data, vis_cube);
					stokesConverter->convert(model_data, model_cube);
					*wtxampDataDivModelS_[vbnum] = amplitude(vis_data) / amplitude(model_data);
				} else {
					*wtxampDataDivModelS_[vbnum] = amplitude(vis_cube) / amplitude(model_cube);
				}
				wtA.reference(*wtxampDataDivModelS_[vbnum]);
				break;
			}
			case PMS::CORR_DIV_MODEL_V: {
				auto corrected_cube = vb->visCubeCorrected();
				auto model_cube = vb->visCubeModel();
				if (stokesConverter) {
					casacore::Cube<casacore::Complex> corrected_data, model_data;
					stokesConverter->convert(corrected_data, corrected_cube);
					stokesConverter->convert(model_data, model_cube);
					*wtxampCorrDivModel_[vbnum] = amplitude(corrected_data / model_data);
				} else {
					*wtxampCorrDivModel_[vbnum] = amplitude(corrected_cube / model_cube);
				}
				wtA.reference(*wtxampCorrDivModel_[vbnum]);
				break;
			}
			case PMS::CORR_DIV_MODEL_S: {
				auto corrected_cube = vb->visCubeCorrected();
				auto model_cube = vb->visCubeModel();
				if (stokesConverter) {
					casacore::Cube<casacore::Complex> corrected_data, model_data;
					stokesConverter->convert(corrected_data, corrected_cube);
					stokesConverter->convert(model_data, model_cube);
					*wtxampCorrDivModelS_[vbnum] = amplitude(corrected_data) / amplitude(model_data);
				} else {
					*wtxampCorrDivModelS_[vbnum] = amplitude(corrected_cube) / amplitude(model_cube);
				}
				wtA.reference(*wtxampCorrDivModelS_[vbnum]);
				break;
			}
			case PMS::FLOAT_DATA: {
				auto float_cube = vb->visCubeFloat();
				if (stokesConverter) {
					casacore::Cube<casacore::Float> float_data;
					stokesConverter->convert(float_data, float_cube);
					*wtxampFloat_[vbnum] = float_data;
				} else {
					*wtxampFloat_[vbnum] = float_cube;
				}
				wtA.reference(*wtxampFloat_[vbnum]);
				break;
			}
		}

		for(uInt c = 0; c < nchannels; ++c) {
			wtA.xzPlane(c) = wtA.xzPlane(c) * weight_data;
		}
		break;
	}
	case PMS::SIGMA: {
		auto sigma = vb->sigma();
		if (stokesConverter) {
			casacore::Array<casacore::Float> sigma_data;
			bool is_sigma(true);
			stokesConverter->convert(sigma_data, sigma, is_sigma);
			*sigma_[vbnum] = sigma_data;
		} else {
			*sigma_[vbnum] = sigma;
		}
		break;
	}
	case PMS::SIGMASP: {
		auto sigmaSpectrum = vb->sigmaSpectrum();
		if (stokesConverter) {
			casacore::Array<casacore::Float> sigmaSpectrum_data;
			bool is_sigma(true);
			stokesConverter->convert(sigmaSpectrum_data, sigmaSpectrum, is_sigma);
			*sigmasp_[vbnum] = sigmaSpectrum_data;
		} else {
			*sigmasp_[vbnum] = sigmaSpectrum;
		}
		break;
	}
	case PMS::AZ0:
	case PMS::EL0: {
		MDirection azelMDir = vb->azel0(vb->time()(0));
		Vector<Double> azelVec = azelMDir.getAngle("deg").getValue();
		az0_(vbnum) = azelVec(0);
		el0_(vbnum) = azelVec(1);
		break;
	}
	case PMS::HA0: {
		ha0_(vbnum) = vb->hourang(vb->time()(0)) * 12 / C::pi;  // in hours
		break;
	}
	case PMS::PA0: {
		pa0_(vbnum) = vb->parang0(vb->time()(0)) * 180.0 / C::pi; // in degrees
		if (pa0_(vbnum) < 0.0) {
			pa0_(vbnum) += 360.0;
		}
		break;
	}
	case PMS::ANTENNA: {
		antenna_[vbnum]->resize(vb->nAntennas());
		indgen(*antenna_[vbnum]);
		break;
	}
	case PMS::AZIMUTH:
	case PMS::ELEVATION: {
		Vector<MDirection> azelVec = vb->azel(vb->time()(0));
		Matrix<Double> azelMat;
				azelMat.resize(2, azelVec.nelements());
		for (uInt iant = 0; iant < azelVec.nelements(); ++iant) {
			azelMat.column(iant) = (azelVec(iant).getAngle("deg").getValue());
		}
		*az_[vbnum] = azelMat.row(0);
		*el_[vbnum] = azelMat.row(1);
		break;
	}
	case PMS::RA:
	case PMS::DEC: {
		auto vbRows = static_cast<size_t>(vb->nRows());
		Vector<MDirection> dir1Vec(vbRows);
		const auto & vbAnt1 = vb->antenna1();
		auto vbTime = vb->time()[0];
		for (decltype(vbRows) k=0; k<vbRows; k++){
			auto antId = vbAnt1[k];
			dir1Vec[k] = (piTvi_->getPointingAngle(antId,vbTime)).second;
		}

		Matrix<Double> raDecMat;
		raDecMat.resize(2, dir1Vec.nelements());
		auto nAnts = dir1Vec.nelements();
		for (decltype(nAnts) iant = 0; iant < nAnts; ++iant) {
			const auto & mVDirection = dir1Vec(iant).getValue();
			const double within_0_360 = 0.0;
			auto dirLong360 = MVAngle(mVDirection.getLong())(within_0_360).degree();
			auto dirLat     = MVAngle(mVDirection.getLat()).degree();
			raDecMat(0,iant) = dirLong360;
			raDecMat(1,iant) = dirLat;
		}
		*((*loadRa_ )[vbnum]) = raDecMat.row(0);
		*((*loadDec_)[vbnum]) = raDecMat.row(1);
		break;
	}
	case PMS::RADIAL_VELOCITY: {
		Int fieldId = vb->fieldId()(0);
		const MSFieldColumns& fieldColumns = vi_p->subtableColumns().field();
		MRadialVelocity radVelocity = fieldColumns.radVelMeas(fieldId, vb->time()(0));
		radialVelocity_(vbnum) = radVelocity.get("AU/d").getValue( "km/s");
		break;
	}
	case PMS::RHO:{
		Int fieldId = vb->fieldId()(0);
		const MSFieldColumns& fieldColumns = vi_p->subtableColumns().field();
		Quantity rhoQuantity = fieldColumns.rho(fieldId, vb->time()(0));
		rho_(vbnum ) = rhoQuantity.getValue( "km");
		break;
	}
	case PMS::PARANG: {
		*parang_[vbnum] = vb->feedPa(vb->time()(0))*(180.0/C::pi);  // in degrees
		break;
	}
	case PMS::ROW: {
		*row_[vbnum] = vb->rowIds();
		break;
	}
	case PMS::OBSERVATION: {
		*obsid_[vbnum] = vb->observationId();
		break;
	}
	case PMS::INTENT:{
		Vector<Int> states = vb->stateId();
		*intent_[vbnum] = assignIntentIds(states);
		break;
	}
	case PMS::FEED1:{
		*feed1_[vbnum] = vb->feed1();
		break;
	}
	case PMS::FEED2:{
		*feed2_[vbnum] = vb->feed2();
		break;
	}
	case PMS::ATM:
	case PMS::TSKY:
	case PMS::IMAGESB: {
		casacore::Int spw = vb->spectralWindows()(0);
		casacore::Int scan = vb->scan()(0);
		casacore::Array<casacore::Double> chanFreqGHz = vb->getFrequencies(0, freqFrame_)/1e9;
		casacore::Vector<casacore::Double> curve(1, 0.0);
		if (axis == PMS::ATM) { 
			if (plotmsAtm_) {
				plotmsAtm_->calcAtmTskyCurve(curve, spw, scan, chanFreqGHz);
			}
			*atm_[vbnum] = curve;
		} else if (axis == PMS::TSKY) {
			if (plotmsAtm_) {
				plotmsAtm_->calcAtmTskyCurve(curve, spw, scan, chanFreqGHz);
			}
			*tsky_[vbnum] = curve;
		} else {
			if (plotmsAtm_) {
				plotmsAtm_->calcImageCurve(curve, spw, scan, chanFreqGHz);
			}
			*imageSideband_[vbnum] = curve;
		}
		break;
	}
	default: {
		throw(AipsError("Axis choice not supported for MS"));
		break;
	}
	} // end switch
}

bool MSCache::isEphemeris(){
	if ( !ephemerisInitialized ){
		Table::TableOption tabopt(Table::Old);
		MeasurementSet ms;
		try {
			ms = MeasurementSet(filename_,TableLock(TableLock::AutoLocking), 
					tabopt);
		} catch (const AipsError& err) {
			throw(AipsError("MeasurementSet setup failed.\n" + err.getMesg()));
		}

		// Check the field subtable for ephemeris fields
		MSColumns msc(ms);
		const MSFieldColumns& fieldColumns = msc.field();
		uInt nrow = fieldColumns.nrow();

		ephemerisAvailable = false;
		String ephemPath;
		for (uInt i=0; i<nrow; ++i) {
			ephemPath = fieldColumns.ephemPath(i);
			if (!ephemPath.empty())
				ephemerisAvailable = true;
		}
		ephemerisInitialized = true;
	}
	return ephemerisAvailable;
}


void MSCache::flagToDisk(const PlotMSFlagging& flagging,
		Vector<Int>& flchunks, Vector<Int>& flrelids,
		Bool setFlag, PlotMSIndexer* indexer, int dataIndex) {

	// Sort the flags by chunk and relative index:
	Sort sorter;
	sorter.sortKey(flchunks.data(),TpInt);
	sorter.sortKey(flrelids.data(),TpInt);

	Vector<uInt> order;  // holds the sort order (indices) for flchunks
	uInt nflag = sorter.sort(order, flchunks.nelements());
	uInt iflag(0);  // index into 'order' array

	bool pmsavg = averaging_.baseline() || averaging_.antenna() || 
		averaging_.spw() || averaging_.scalarAve();
	// check if extending flags and subset selected
	bool extendcorr = flagging.corr() && !selection_.corr().empty();
	bool extendchan = flagging.channel() && !selection_.spw().empty();

	// get loadedAxes and loadedData for setting up vis iter
	vector<PMS::Axis> loadedAxes;
	vector<PMS::DataColumn> loadedData;
	for (std::map<PMS::Axis, bool>::iterator mapit=loadedAxes_.begin(); 
			mapit!=loadedAxes_.end(); ++mapit) {
		if (mapit->second) {
			PMS::Axis loadedAxis = mapit->first;
			if (loadedAxesData_.find(loadedAxis) != loadedAxesData_.end()) {
				uInt ndatacols = loadedAxesData_[loadedAxis].nfields();
				for (uInt i=0; i<ndatacols; ++i) {
					loadedAxes.push_back(loadedAxis);
					String datacolStr = loadedAxesData_[loadedAxis].name(i);
					loadedData.push_back(PMS::dataColumn(datacolStr));
				}
			} else {
				loadedAxes.push_back(loadedAxis);
				loadedData.push_back(PMS::DATA);
			}
		}
	}

	// If extending flags, select all correlations or channels in vis iter
	PlotMSSelection flagSel = selection_;
	if (extendcorr) {
		// select all corrs
		flagSel.setCorr("");
	}
	if (extendchan) {
		// select spws only (all channels per spw)
		MeasurementSet ms(filename_);
		MSSelection mssel(ms);
		mssel.setSpwExpr(selection_.spw());
		Vector<Int> spws = mssel.getSpwList();
		String spwExpr = casacore::String::toString(spws(0));
		for (uInt spw=1; spw<spws.size(); ++spw) 
			spwExpr += "," + casacore::String::toString(spws(spw));
		flagSel.setSpw(spwExpr);
	}

	setUpVisIter(flagSel, calibration_, dataColumn_, loadedAxes,
		loadedData, true, false);
	vi_p->originChunks();
	vi_p->origin();
	vi::VisBuffer2* vb = vi_p->getVisBuffer();

	// iterate through chunks and find the ones that need flags changed
	for (Int ichk=0; ichk<nChunk_; ++ichk) {
		if (ichk != flchunks(order[iflag])) {
			// Step over current chunk if not in flchunks
			for (Int i=0;i<nVBPerAve_(ichk);++i) {
				vi_p->next();
				if (!vi_p->more() && vi_p->moreChunks()) {
					vi_p->nextChunk();
					vi_p->origin();
				}
			}
		} else if ( pmsavg || extendcorr || extendchan) {
			// This chunk requires flag-setting but have to handle chunks
			// (shape has changed: VBs averaged together or extending flags)
			Cube<Bool> vbflag;
			Vector<Bool> vbflagrow;
			Vector<Int> a1, a2;

			for (Int i=0; i<nVBPerAve_(ichk); ++i) {

				// Refer to VB pieces we need
				vbflag = vb->flagCube();
				vbflagrow = vb->flagRow();
				a1 = vb->antenna1();
				a2 = vb->antenna2();
				Int ncorr = vb->nCorrelations();
				Int nchan = vb->nChannels();
				Int nrow  = vb->nRows();

				// Apply all flags in this chunk to this VB
				while (iflag<nflag && flchunks(order[iflag])==ichk) {

					Int currChunk = flchunks(order[iflag]);
					Int irel = flrelids(order[iflag]);
					Slice corr,chan,bsln;

					// Set flag range on correlation axis:
					if (netAxesMask_[dataIndex](0) && !flagging.corrAll()) {
						// A specific single correlation
						Int icorr = indexer->getIndex1000(currChunk, irel);
						corr = Slice(icorr, 1, 1);
					} else {
						// Extend to all correlations
						corr=Slice(0, ncorr, 1);
					}

					// Set Flag range on channel axis:
					if (netAxesMask_[dataIndex](1) && !flagging.channel()) {
						// A single specific channel
						Int ichan = indexer->getIndex0100(currChunk, irel);
						chan = Slice(ichan, 1, 1);
					} else {
						// Extend to all channels
						chan = Slice(0, nchan, 1);
					}

					// Set Flags on the baseline axis:
					Int thisAnt1 = Int(getAnt1(currChunk,
						indexer->getIndex0010(currChunk, irel)));
					Int thisAnt2 = Int(getAnt2(currChunk,
						indexer->getIndex0010(currChunk, irel)));
					if (netAxesMask_[dataIndex](2) &&
						!flagging.antennaBaselinesBased() && (thisAnt1 > -1) ) {
						// i.e., if baseline is an explicit data axis,
						//	   full baseline extension is OFF
						//	   and the first antenna in the selected point is > -1
						// Do some variety of detailed per-baseline flagging
						for (Int irow=0; irow<nrow; ++irow) {
							if (thisAnt2 > -1) {
								// match a baseline exactly
								if ((a1(irow)==thisAnt1) &&
									(a2(irow)==thisAnt2)) {
									vbflag(corr, chan, Slice(irow,1,1)) = setFlag;
									// unset flag_row when unflagging
									if (!setFlag) vbflagrow(irow) = false;
									break;  // found the one baseline, escape from for loop
								}
							} else {
								// either antenna matches the one specified antenna
								//  (don't break because there will be more than one)
								//  TBD: this doesn't get cross-hands quite right when
								//	averaging 'by antenna'...
								if ((a1(irow) == thisAnt1) ||
									(a2(irow)==thisAnt1)) {
									vbflag(corr, chan, Slice(irow,1,1)) = setFlag;
									// unset flag_row when unflagging
									if (!setFlag) vbflagrow(irow) = false;
								}
							}
						}
					} else {
						// Set flags for all baselines, because the plot
						//  is ordinarily implicit in baseline, we've turned on baseline
						//  extension, or we've averaged over all baselines
						bsln = Slice(0, nrow, 1);
						vbflag(corr, chan, bsln) = setFlag;
						// unset flag_row when unflagging
						if (!setFlag) vbflagrow(bsln) = false;
					}
					++iflag;
				} // done with this flchunk

				// Put the flags back into the MS
				vi_p->writeFlag(vbflag);
				// Advance to the next vb
				vi_p->next();
				if (!vi_p->more() && vi_p->moreChunks()) {
					vi_p->nextChunk();
					vi_p->origin();
				}
			}  // VBs in this averaging chunk

			// Escape if we are already finished
			if (iflag >= nflag) break;

		} else {
			// same shape (possibly averaging handled by MSTransformIterator)
			// just write out flag chunk
			vi_p->writeFlag(flag(ichk));

			// Advance to the next vb
			vi_p->next();
			if (!vi_p->more() && vi_p->moreChunks()) {
				vi_p->nextChunk();
				vi_p->origin();
			}
			// Advance to the next flagged chunk
			++iflag;
			while ((iflag < nflag) && 
				   (flchunks(order[iflag]) == flchunks(order[iflag-1]))) {
				++iflag;
			}

			// Escape if we are finished
			if (iflag >= nflag) break;
		} 
	}
	// Delete the VisIter so lock is released
	deleteVi();
}

void MSCache::mapIntentNamesToIds() {
	intentIds_.clear();
	Int newId = 0;

	for (uInt i=0; i<intentnames_.size(); i++) {
		if (intentIds_.find(intentnames_[i]) == intentIds_.end()) {
			intentIds_[intentnames_[i]] = newId;
			newId++;
		}
	}
}

Vector<Int> MSCache::assignIntentIds(Vector<Int>& stateIds) {
	Vector<Int> intents;
	intents.resize(stateIds.size());

	for (uInt i=0; i<stateIds.size(); i++) {
		if ((intentnames_.size() > 0) && (stateIds[i] >= 0))
			intents[i] = intentIds_[intentnames_[stateIds[i]]];
		else
			intents[i] = stateIds[i];
	}

	return intents;
}

Vector<Double> MSCache::calcVelocity(vi::VisBuffer2* vb) {
	// Convert freq in the vb to velocity
	Vector<Double> outVel;
	VisBufferUtil vbu = VisBufferUtil();
	vbu.toVelocity(outVel,
			*vb,
			freqFrame_,
			MVFrequency(transformations_.restFreqHz()),
			transformations_.veldef());
	outVel /= 1.0e3;  // in km/s
	return outVel;
}

bool MSCache::pointingsPlotSupported(const MeasurementSet* const &ms){
	// MS must have a non empty pointing table
	const auto & pointingTable = ms->pointing();
	if (pointingTable.nrow() == 0 ) {
		logWarn(PMS::LOG_ORIGIN_LOAD_CACHE, "Plotting antennas directions is irrelevant when the pointing table is empty.");
		return false;
	}

	// Telescope must have been validated for this feature (CAS-8087)
	Float noMetaDataCache = -1.0;
	MSMetaData msmd(ms,noMetaDataCache);
	vector<String> observatoryNames {msmd.getObservatoryNames()};
	bool isValidated = false;
	for (const auto & observatoryName : observatoryNames) {
		if (	observatoryName == "ALMA" ||
				observatoryName == "ASTE" ||
				observatoryName == "NRO"
			)
		{
			isValidated = true;
			continue;
		}
		else {
			isValidated = false;
			String warnMsg;
			if (observatoryName == "EVLA" ) {
				warnMsg  = "Plotting " + observatoryName + "'s antennas directions requires telescope-specific processing,\n";
				warnMsg += "which is currently not implemented";
			}
			else {
				warnMsg = "Plotting antennas directions has not been validated for observatory: ";
				warnMsg += observatoryName;
			}
			logWarn(PMS::LOG_ORIGIN_LOAD_CACHE, warnMsg);
			break;
		}
	}
	return isValidated;
}

void MSCache::loadPageHeaderCache(const casacore::MeasurementSet& selectedMS){
	logLoad("Loading page header cache");
	using Item = PageHeaderItemsDef::Item;

	// ---- Filename
	pageHeaderCache_.store(HeaderItemData(Item::Filename,Path(filename_).baseName()));

	if ( selectedMS.nrow() == 0 ) return;

	const uInt firstSelectedRow = 0;

	MSColumns selMSColumns(selectedMS);

	// ---- Queries on Observation table
	auto firstObservationId = selMSColumns.observationId().get(firstSelectedRow);
	auto firstObservationRow = static_cast<uInt>(firstObservationId);
	const auto & obsColumns = selMSColumns.observation();

	if ( firstObservationRow < obsColumns.nrow() ) {

		// ---- Observation Start Date
		auto timeRangeMeasure = obsColumns.timeRangeMeas()(firstObservationRow);
		IPosition startTimePos(1,0);
		auto obsStartMEpoch = timeRangeMeasure(startTimePos);

		Bool isUTC = (obsStartMEpoch.getRef().getType() == MEpoch::UTC);
		if ( ! isUTC ) {
			obsStartMEpoch = MEpoch::Convert(obsStartMEpoch,MEpoch::Ref(MEpoch::UTC))();
		}

		MVTime obsStartMVTime(obsStartMEpoch.getValue());

		stringstream obsStartDateStream;
		obsStartDateStream << obsStartMVTime.monthday()  << " "
						   << obsStartMVTime.monthName() << " "
						   << obsStartMVTime.year();
		auto obsStartDate = obsStartDateStream.str();
		pageHeaderCache_.store(HeaderItemData(Item::Obs_Start_Date,obsStartDate));

		// ---- Observation Start Time
		const uInt timePrecision = 4;
		MVTime::Format timeFormat(MVTime::TIME | MVTime::CLEAN,timePrecision);
		auto obsStartTime = obsStartMVTime.string(timeFormat);
		obsStartTime += String(" ") + MEpoch::showType(obsStartMEpoch.getRef().getType());
		pageHeaderCache_.store(HeaderItemData(Item::Obs_Start_Time,obsStartTime));

		// ---- Observer
		auto observer = obsColumns.observer().get(firstObservationRow);
		pageHeaderCache_.store(HeaderItemData(Item::Obs_Observer,observer));
		// ---- Project
		auto project = obsColumns.project().get(firstObservationRow);
		pageHeaderCache_.store(HeaderItemData(Item::Obs_Project,project));
		// ---- Telescope Name
		auto telescopeName = obsColumns.telescopeName().get(firstObservationRow);
		pageHeaderCache_.store(HeaderItemData(Item::Obs_Telescope_Name,telescopeName));
	}
	// Target Name
	// Lookup NAME of first selected row in FIELD table
	auto firstFieldId = selMSColumns.fieldId().get(firstSelectedRow);
	auto firstFieldRow = static_cast<uInt>(firstFieldId);
	const auto & fieldColumns = selMSColumns.field();
	auto firstFieldName = fieldColumns.name().get(firstFieldRow);
	pageHeaderCache_.store(HeaderItemData(Item::Target_Name,firstFieldName));

	// Target Direction
	auto haveSourceTable = ! selectedMS.source().isNull();
	if ( haveSourceTable ) {
		// Lookup SOURCE_ID of first selected row in FIELD table
		auto firstSourceId = fieldColumns.sourceId().get(firstFieldRow);
		// Lookup SPECTRAL_WINDOW_ID of first selected row in DATA_DESCRIPTION table
		auto firstDataDescId = selMSColumns.dataDescId().get(firstSelectedRow);
		auto firstDataDescRow = static_cast<uInt>(firstDataDescId);
		const auto & dataDescColumns = selMSColumns.dataDescription();
		auto firstSpwId = dataDescColumns.spectralWindowId().get(firstDataDescRow);
		// Query target direction from SOURCE table
		auto & sourceTable = selectedMS.source();
		// ---- Shorthands
		using SourceColumn = MSSourceEnums::PredefinedColumns;
		auto getTen = [sourceTable](SourceColumn colEnum) {
			return sourceTable.col(sourceTable.columnName(colEnum)) ;
		};
		// ---- Table Expression Nodes
		auto tenSourceID = getTen(SourceColumn::SOURCE_ID);
		auto tenSpwID = getTen(SourceColumn::SPECTRAL_WINDOW_ID);
		// ---- Selection criteria
		auto sourceMatch { tenSourceID == firstSourceId };
		auto spwIdAny   { tenSpwID == -1  /* row valid for any SPECTRAL_WINDOW_ID */  };
		auto spwIdMatch { tenSpwID == firstSpwId };
		// ---- Select and sort by ascending TIME
		auto qryTargetDirection { sourceMatch && (spwIdAny || spwIdMatch) };
		const auto & timeColName = sourceTable.columnName(SourceColumn::TIME);
		Table qryResult = sourceTable(qryTargetDirection).sort(timeColName,Sort::Ascending);
		auto qryResultRows = qryResult.rowNumbers(sourceTable);
		// ---- Pick-up first direction
		if ( ! qryResultRows.empty() ) {
			auto firstDirectionRow = qryResultRows[0];
			const auto & sourceColumns = selMSColumns.source();
			auto firstDirectionMeasure = sourceColumns.directionMeas()(firstDirectionRow);
			const auto & dir  = firstDirectionMeasure;
			// auto sourceDirection = firstDirectionMeasure.toString();
			auto refType = MDirection::castType(dir.getRef().getType());
			String sourceDirection;
			if ( refType == MDirection::Types::ICRS) {
				// Format the same way casacore::MDirection::toString() formats when refType == J2000
				auto longitude = dir.getValue().getLong("deg");
				auto lat = dir.getValue().getLat("deg");
				auto ra = MVTime(longitude).string(MVTime::TIME, 12);
				auto dec = MVAngle(abs(lat)).string(MVAngle::ANGLE_CLEAN, 11);
				dec.trim();
				if ( lat.getValue() < 0) {
					dec = "-" + dec;
				}
				sourceDirection = ra + " " + dec;
			}
			else sourceDirection = firstDirectionMeasure.toString();

			pageHeaderCache_.store(HeaderItemData(Item::Target_Direction,sourceDirection));
		}

	}
}

void MSCache::completeLoadPageHeaderCache(){
	using Item = PageHeaderItemsDef::Item;
	// ---- Y Column(s)
	String yColumns;
	String axisSep;
	for ( uInt k=0; k<currentY_.size(); k++) {
		yColumns += axisSep;
		auto axis = currentY_[k];
		yColumns += PMS::axis(axis);
		if ( PMS::axisIsData(axis) && k < currentYData_.size() ) {
			auto dataColumn = currentYData_[k];
			yColumns += String(":") + PMS::dataColumn(dataColumn);
		}
		axisSep = String(", ");
	}
	pageHeaderCache_.store(HeaderItemData(Item::YColumns,yColumns));
}

}
